################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.datastream import StreamExecutionEnvironment
from pyflink.common import (ExecutionConfig, Configuration)
from pyflink.java_gateway import get_gateway
from pyflink.testing.test_case_utils import PyFlinkTestCase
from pyflink.util.java_utils import get_j_env_configuration


class ExecutionConfigTests(PyFlinkTestCase):

    def setUp(self):
        self.env = StreamExecutionEnvironment.get_execution_environment()
        self.execution_config = self.env.get_config()

    def test_constant(self):
        gateway = get_gateway()
        JExecutionConfig = gateway.jvm.org.apache.flink.api.common.ExecutionConfig

        self.assertEqual(ExecutionConfig.PARALLELISM_DEFAULT, JExecutionConfig.PARALLELISM_DEFAULT)
        self.assertEqual(ExecutionConfig.PARALLELISM_UNKNOWN, JExecutionConfig.PARALLELISM_UNKNOWN)

    def test_get_set_closure_cleaner(self):

        self.assertTrue(self.execution_config.is_closure_cleaner_enabled())

        self.execution_config.disable_closure_cleaner()

        self.assertFalse(self.execution_config.is_closure_cleaner_enabled())

        self.execution_config.enable_closure_cleaner()

        self.assertTrue(self.execution_config.is_closure_cleaner_enabled())

    def test_get_set_auto_watermark_interval(self):

        self.assertEqual(self.execution_config.get_auto_watermark_interval(), 200)

        self.execution_config.set_auto_watermark_interval(1000)

        self.assertEqual(self.execution_config.get_auto_watermark_interval(), 1000)

    def test_get_set_parallelism(self):

        self.execution_config.set_parallelism(8)

        self.assertEqual(self.execution_config.get_parallelism(), 8)

        self.execution_config.set_parallelism(4)

        self.assertEqual(self.execution_config.get_parallelism(), 4)

    def test_get_set_max_parallelism(self):

        self.execution_config.set_max_parallelism(12)

        self.assertEqual(self.execution_config.get_max_parallelism(), 12)

        self.execution_config.set_max_parallelism(16)

        self.assertEqual(self.execution_config.get_max_parallelism(), 16)

    def test_get_set_task_cancellation_interval(self):

        self.assertEqual(self.execution_config.get_task_cancellation_interval(), 30000)

        self.execution_config.set_task_cancellation_interval(1000)

        self.assertEqual(self.execution_config.get_task_cancellation_interval(), 1000)

    def test_get_set_task_cancellation_timeout(self):

        self.assertEqual(self.execution_config.get_task_cancellation_timeout(), 180000)

        self.execution_config.set_task_cancellation_timeout(3000)

        self.assertEqual(self.execution_config.get_task_cancellation_timeout(), 3000)

    def test_disable_enable_auto_generated_uids(self):

        self.execution_config.disable_auto_generated_uids()

        self.assertFalse(self.execution_config.has_auto_generated_uids_enabled())

        self.execution_config.enable_auto_generated_uids()

        self.assertTrue(self.execution_config.has_auto_generated_uids_enabled())

    def test_disable_enable_object_reuse(self):

        self.execution_config.disable_object_reuse()

        self.assertFalse(self.execution_config.is_object_reuse_enabled())

        self.execution_config.enable_object_reuse()

        self.assertTrue(self.execution_config.is_object_reuse_enabled())

    def test_get_set_global_job_parameters(self):

        self.execution_config.set_global_job_parameters({"hello": "world"})

        self.assertEqual(self.execution_config.get_global_job_parameters(), {"hello": "world"})

    def test_get_set_use_snapshot_compression(self):

        self.execution_config.set_use_snapshot_compression(False)

        self.assertFalse(self.execution_config.is_use_snapshot_compression())

        self.execution_config.set_use_snapshot_compression(True)

        self.assertTrue(self.execution_config.is_use_snapshot_compression())

    def test_equals_and_hash(self):

        config1 = StreamExecutionEnvironment.get_execution_environment().get_config()

        config2 = StreamExecutionEnvironment.get_execution_environment().get_config()

        self.assertEqual(config1, config2)

        self.assertEqual(hash(config1), hash(config2))

        config1.set_parallelism(12)
        config2.set_parallelism(11)

        self.assertNotEqual(config1, config2)

        # it is allowed for hashes to be equal even if objects are not

        config2.set_parallelism(12)

        self.assertEqual(config1, config2)

        self.assertEqual(hash(config1), hash(config2))

    def test_get_execution_environment_with_config(self):
        configuration = Configuration()
        configuration.set_integer('parallelism.default', 12)
        configuration.set_string('pipeline.name', 'haha')
        env = StreamExecutionEnvironment.get_execution_environment(configuration)
        execution_config = env.get_config()

        self.assertEqual(execution_config.get_parallelism(), 12)
        config = Configuration(
            j_configuration=get_j_env_configuration(env._j_stream_execution_environment))
        self.assertEqual(config.get_string('pipeline.name', ''), 'haha')



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from pyflink.common.utils import JavaObjectWrapper
from pyflink.java_gateway import get_gateway

__all__ = [
    'SerializationSchema',
    'DeserializationSchema',
    'SimpleStringSchema',
    'ByteArraySchema',
    'Encoder',
    'BulkWriterFactory',
    'RowFieldExtractorSchema',
]


class SerializationSchema(object):
    """
    Base class for SerializationSchema. The serialization schema describes how to turn a data object
    into a different serialized representation. Most data sinks (for example Apache Kafka) require
    the data to be handed to them in a specific format (for example as byte strings).
    """

    def __init__(self, j_serialization_schema=None):
        self._j_serialization_schema = j_serialization_schema


class DeserializationSchema(object):
    """
    Base class for DeserializationSchema. The deserialization schema describes how to turn the byte
    messages delivered by certain data sources (for example Apache Kafka) into data types (Java/
    Scala objects) that are processed by Flink.

    In addition, the DeserializationSchema describes the produced type which lets Flink create
    internal serializers and structures to handle the type.
    """

    def __init__(self, j_deserialization_schema=None):
        self._j_deserialization_schema = j_deserialization_schema


class SimpleStringSchema(SerializationSchema, DeserializationSchema):
    """
    Very simple serialization/deserialization schema for strings. By default, the serializer uses
    'UTF-8' for string/byte conversion.
    """

    def __init__(self, charset: str = 'UTF-8'):
        gate_way = get_gateway()
        j_char_set = gate_way.jvm.java.nio.charset.Charset.forName(charset)
        j_simple_string_serialization_schema = gate_way \
            .jvm.org.apache.flink.api.common.serialization.SimpleStringSchema(j_char_set)
        SerializationSchema.__init__(self,
                                     j_serialization_schema=j_simple_string_serialization_schema)
        DeserializationSchema.__init__(
            self, j_deserialization_schema=j_simple_string_serialization_schema)


class ByteArraySchema(SerializationSchema, DeserializationSchema):
    """
    Simple serialization/deserialization schema for bytes.
    """

    def __init__(self):
        gate_way = get_gateway()
        j_simple_byte_serialization_schema = gate_way \
            .jvm.org.apache.flink.api.common.serialization.ByteArraySchema()
        SerializationSchema.__init__(self,
                                     j_serialization_schema=j_simple_byte_serialization_schema)
        DeserializationSchema.__init__(
            self, j_deserialization_schema=j_simple_byte_serialization_schema)


class Encoder(object):
    """
    Encoder is used by the file sink to perform the actual writing of the
    incoming elements to the files in a bucket.
    """

    def __init__(self, j_encoder):
        self._j_encoder = j_encoder

    @staticmethod
    def simple_string_encoder(charset_name: str = "UTF-8") -> 'Encoder':
        """
        A simple Encoder that uses toString() on the input elements and writes them to
        the output bucket file separated by newline.
        """
        j_encoder = get_gateway().jvm.org.apache.flink.api.common.serialization.\
            SimpleStringEncoder(charset_name)
        return Encoder(j_encoder)


class BulkWriterFactory(JavaObjectWrapper):
    """
    The Python wrapper of Java BulkWriter.Factory interface, which is the base interface for data
    sinks that write records into files in a bulk manner.
    """

    def __init__(self, j_bulk_writer_factory):
        super().__init__(j_bulk_writer_factory)


class RowDataBulkWriterFactory(BulkWriterFactory):
    """
    A :class:`～BulkWriterFactory` that receives records with RowData type. This is for indicating
    that Row record from Python must be first converted to RowData.
    """

    def __init__(self, j_bulk_writer_factory, row_type):
        super().__init__(j_bulk_writer_factory)
        self._row_type = row_type

    def get_row_type(self):
        return self._row_type


class RowFieldExtractorSchema(SerializationSchema):
    """
    Serialization schema that extracts a specific field from a Row and returns it as a
    byte array. The field at the specified index MUST be of type bytes (byte array).
    This schema is particularly useful when using Flink with Kafka, where you may want to use a
    specific field as the message key for partition routing.
    The field being extracted must already be a byte array. Users are responsible for
    converting their data to bytes before passing it to this schema.

    Example usage with Kafka:
        >>> from pyflink.common.serialization import RowFieldExtractorSchema
        >>> from pyflink.datastream.connectors.kafka import KafkaSink, \
            KafkaRecordSerializationSchema
        >>>
        >>> # User must convert data to bytes beforehand
        >>> # For example: Row.of(b"key-bytes", b"value-bytes")
        >>>
        >>> sink = KafkaSink.builder() \\
        ...     .set_bootstrap_servers("localhost:9092") \\
        ...     .set_record_serializer(
        ...         KafkaRecordSerializationSchema.builder()
        ...             .set_topic("my-topic")
        ...             .set_key_serialization_schema(RowFieldExtractorSchema(0))
                        # Field 0 (must be bytes) as key
        ...             .set_value_serialization_schema(RowFieldExtractorSchema(1))
                        # Field 1 (must be bytes) as value
        ...             .build()
        ...     ) \\
        ...     .build()

    :param field_index: The zero-based index of the field to extract from the Row.
    The field at this index must be of type bytes.
    """

    def __init__(self, field_index: int):
        """
        Creates a new RowFieldExtractorSchema that extracts the field at the specified index.

        :param field_index: The zero-based index of the field to extract (must be non-negative).
        :raises ValueError: If field_index is negative.
        """
        if field_index < 0:
            raise ValueError(f"Field index must be non-negative, got: {field_index}")
        gateway = get_gateway()
        j_row_field_extractor_schema = gateway.jvm.org.apache.flink.api.common.serialization \
            .RowFieldExtractorSchema(field_index)
        super(RowFieldExtractorSchema, self).__init__(
            j_serialization_schema=j_row_field_extractor_schema)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.table import StatementSet
from pyflink.testing.test_case_utils import PythonAPICompletenessTestCase, PyFlinkTestCase


class StatementSetAPICompletenessTests(PythonAPICompletenessTestCase, PyFlinkTestCase):
    """
    Tests whether the Python :class:`StatementSet` is consistent with
    Java `org.apache.flink.table.api.StatementSet`.
    """

    @classmethod
    def python_class(cls):
        return StatementSet

    @classmethod
    def java_class(cls):
        return "org.apache.flink.table.api.StatementSet"


if __name__ == "__main__":
    import unittest

    try:
        import xmlrunner

        testRunner = xmlrunner.XMLTestRunner(output="target/test-reports")
    except ImportError:
        testRunner = None
    unittest.main(testRunner=testRunner, verbosity=2)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from py4j.protocol import Py4JJavaError

from pyflink.table import expressions as expr
from pyflink.table.window import Session, Slide, Tumble, Over
from pyflink.testing.test_case_utils import PyFlinkStreamTableTestCase, PyFlinkTestCase


class StreamTableWindowTests(PyFlinkStreamTableTestCase):

    def test_over_window(self):
        t_env = self.t_env
        t = t_env.from_elements([(1, 1, "Hello")], ['a', 'b', 'c'])

        result = t.over_window(
            Over.partition_by(t.c)
                .order_by(t.a)
                .preceding(expr.row_interval(2))
                .following(expr.CURRENT_ROW)
                .alias("w"))

        self.assertRaisesRegex(
            Py4JJavaError,
            "Second argument in OVER window must be a TIME ATTRIBUTE, but is: BIGINT",
            result.select, expr.col("b").sum.over(expr.col("w")))


class BatchTableWindowTests(PyFlinkTestCase):

    def setUp(self):
        from pyflink.table import TableEnvironment
        from pyflink.table import EnvironmentSettings
        self.t_env = TableEnvironment.create(EnvironmentSettings.in_batch_mode())

    def test_tumble_window(self):
        t = self.t_env.from_elements([(1, 1, "Hello")], ["a", "b", "c"])
        result = t.window(Tumble.over(expr.row_interval(2)).on(expr.col("a")).alias("w"))\
            .group_by(expr.col('w'), expr.col('c')).select(t.b.sum)

        query_operation = result._j_table.getQueryOperation().getChildren().get(0)
        self.assertEqual('[c]', query_operation.getGroupingExpressions().toString())
        self.assertEqual('TumbleWindow(field: [a], size: [2])',
                         query_operation.getGroupWindow().asSummaryString())

    def test_slide_window(self):
        t = self.t_env.from_elements([(1000, 1, "Hello")], ["a", "b", "c"])
        result = t.window(Slide.over(expr.lit(2).seconds).every(expr.lit(1).seconds).on(t.a)
                          .alias("w")).group_by(expr.col('w'), expr.col('c')).select(t.b.sum)

        query_operation = result._j_table.getQueryOperation().getChildren().get(0)
        self.assertEqual('[c]', query_operation.getGroupingExpressions().toString())
        self.assertEqual('SlideWindow(field: [a], slide: [1000], size: [2000])',
                         query_operation.getGroupWindow().asSummaryString())

    def test_session_window(self):
        t = self.t_env.from_elements([(1000, 1, "Hello")], ["a", "b", "c"])
        result = t.window(Session.with_gap(expr.lit(1).seconds).on(t.a).alias("w"))\
            .group_by(expr.col('w'), expr.col('c')).select(t.b.sum)

        query_operation = result._j_table.getQueryOperation().getChildren().get(0)
        self.assertEqual('[c]', query_operation.getGroupingExpressions().toString())
        self.assertEqual('SessionWindow(field: [a], gap: [1000])',
                         query_operation.getGroupWindow().asSummaryString())


if __name__ == '__main__':
    import unittest

    try:
        import xmlrunner
        testRunner = xmlrunner.XMLTestRunner(output='target/test-reports')
    except ImportError:
        testRunner = None
    unittest.main(testRunner=testRunner, verbosity=2)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
# #  distributed under the License is distributed on an "AS IS" BASIS,
# #  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# #  See the License for the specific language governing permissions and
# # limitations under the License.
################################################################################
from pyflink.table import TableEnvironment, EnvironmentSettings
from pyflink.testing.test_case_utils import PyFlinkTestCase


class StreamTableSetOperationTests(PyFlinkTestCase):

    data1 = [(1, "Hi", "Hello")]
    data2 = [(3, "Hello", "Hello")]
    schema = ["a", "b", "c"]

    def setUp(self) -> None:
        self.t_env = TableEnvironment.create(EnvironmentSettings.in_batch_mode())

    def test_minus(self):
        t_env = self.t_env
        t1 = t_env.from_elements(self.data1, self.schema)
        t2 = t_env.from_elements(self.data2, self.schema)

        result = t1.minus(t2)

        self.assertEqual('MINUS', result._j_table.getQueryOperation().getType().toString())
        self.assertFalse(result._j_table.getQueryOperation().isAll())

    def test_minus_all(self):
        t_env = self.t_env
        t1 = t_env.from_elements(self.data1, self.schema)
        t2 = t_env.from_elements(self.data2, self.schema)

        result = t1.minus_all(t2)
        self.assertEqual('MINUS', result._j_table.getQueryOperation().getType().toString())
        self.assertTrue(result._j_table.getQueryOperation().isAll())

    def test_union(self):
        t_env = self.t_env
        t1 = t_env.from_elements(self.data1, self.schema)
        t2 = t_env.from_elements(self.data2, self.schema)

        result = t1.union(t2)
        self.assertEqual('UNION', result._j_table.getQueryOperation().getType().toString())
        self.assertFalse(result._j_table.getQueryOperation().isAll())

    def test_union_all(self):
        t_env = self.t_env
        t1 = t_env.from_elements(self.data1, self.schema)
        t2 = t_env.from_elements(self.data2, self.schema)

        result = t1.union_all(t2)
        self.assertEqual('UNION', result._j_table.getQueryOperation().getType().toString())
        self.assertTrue(result._j_table.getQueryOperation().isAll())

    def test_intersect(self):
        t_env = self.t_env
        t1 = t_env.from_elements(self.data1, self.schema)
        t2 = t_env.from_elements(self.data2, self.schema)

        result = t1.intersect(t2)
        self.assertEqual('INTERSECT', result._j_table.getQueryOperation().getType().toString())
        self.assertFalse(result._j_table.getQueryOperation().isAll())

    def test_intersect_all(self):
        t_env = self.t_env
        t1 = t_env.from_elements(self.data1, self.schema)
        t2 = t_env.from_elements(self.data2, self.schema)

        result = t1.intersect_all(t2)
        self.assertEqual('INTERSECT', result._j_table.getQueryOperation().getType().toString())
        self.assertTrue(result._j_table.getQueryOperation().isAll())


if __name__ == '__main__':
    import unittest

    try:
        import xmlrunner
        testRunner = xmlrunner.XMLTestRunner(output='target/test-reports')
    except ImportError:
        testRunner = None
    unittest.main(testRunner=testRunner, verbosity=2)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from typing import Dict, Union, Optional

from pyflink.common.config_options import ConfigOption
from pyflink.java_gateway import get_gateway
from pyflink.table.schema import Schema

__all__ = ['ModelDescriptor']


class ModelDescriptor(object):
    """
    Describes a CatalogModel representing a model.

    ModelDescriptor is a template for creating a CatalogModel instance. It closely resembles the
    "CREATE MODEL" SQL DDL statement, containing input schema, output schema, and other
    characteristics.

    This can be used to register a model in the Table API.
    """
    def __init__(self, j_model_descriptor):
        self._j_model_descriptor = j_model_descriptor

    @staticmethod
    def for_provider(provider: str) -> 'ModelDescriptor.Builder':
        """
        Creates a new :class:`~pyflink.table.ModelDescriptor.Builder` for a model using the given
        provider value.

        :param provider: The provider value for the model.
        """
        gateway = get_gateway()
        j_builder = gateway.jvm.ModelDescriptor.forProvider(provider)
        return ModelDescriptor.Builder(j_builder)

    def get_options(self) -> Dict[str, str]:
        return self._j_model_descriptor.getOptions()

    def get_input_schema(self) -> Optional[Schema]:
        j_input_schema = self._j_model_descriptor.getInputSchema()
        if j_input_schema.isPresent():
            return Schema(j_input_schema.get())
        else:
            return None

    def get_output_schema(self) -> Optional[Schema]:
        j_output_schema = self._j_model_descriptor.getOutputSchema()
        if j_output_schema.isPresent():
            return Schema(j_output_schema.get())
        else:
            return None

    def get_comment(self) -> Optional[str]:
        j_comment = self._j_model_descriptor.getComment()
        if j_comment.isPresent():
            return j_comment.get()
        else:
            return None

    def __str__(self):
        return self._j_model_descriptor.toString()

    def __eq__(self, other):
        return (self.__class__ == other.__class__ and
                self._j_model_descriptor.equals(other._j_model_descriptor))

    def __hash__(self):
        return self._j_model_descriptor.hashCode()

    class Builder(object):
        """
        Builder for ModelDescriptor.
        """

        def __init__(self, j_builder):
            self._j_builder = j_builder

        def input_schema(self, input_schema: Schema) -> 'ModelDescriptor.Builder':
            """
            Define the input schema of the ModelDescriptor.
            """
            self._j_builder.inputSchema(input_schema._j_schema)
            return self

        def output_schema(self, output_schema: Schema) -> 'ModelDescriptor.Builder':
            """
            Define the output schema of the ModelDescriptor.
            """
            self._j_builder.outputSchema(output_schema._j_schema)
            return self

        def option(self, key: Union[str, ConfigOption], value) -> 'ModelDescriptor.Builder':
            """
            Sets the given option on the model.

            Option keys must be fully specified.

            Example:
            ::

                >>> ModelDescriptor.for_connector("OPENAI")\
                ...     .input_schema(input_schema)\
                ...     .output_schema(output_schema)\
                ...     .option("task", "regression")\
                ...     .build()

            """
            if isinstance(key, str):
                self._j_builder.option(key, value)
            else:
                self._j_builder.option(key._j_config_option, value)
            return self

        def comment(self, comment: str) -> 'ModelDescriptor.Builder':
            """
            Define the comment for this model.
            """
            self._j_builder.comment(comment)
            return self

        def build(self) -> 'ModelDescriptor':
            """
            Returns an immutable instance of :class:`~pyflink.table.ModelDescriptor`.
            """
            return ModelDescriptor(self._j_builder.build())



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.table import DataTypes, ResultKind
from pyflink.table.schema import Schema
from pyflink.table.table_descriptor import TableDescriptor
from pyflink.testing.test_case_utils import PyFlinkStreamTableTestCase


class TablePipelineTest(PyFlinkStreamTableTestCase):

    def test_table_pipeline_with_anonymous_sink(self):
        schema = Schema.new_builder().column("f0", DataTypes.STRING()).build()
        table = self.t_env.from_descriptor(
            TableDescriptor.for_connector("datagen")
            .option("number-of-rows", "10")
            .schema(schema)
            .build()
        )
        table_pipeline = table.insert_into(
            TableDescriptor.for_connector("blackhole").schema(schema).build()
        )
        self.assertEqual(table_pipeline.get_sink_identifier(), None)

    def test_table_pipeline_with_registered_sink(self):
        schema = Schema.new_builder().column("f0", DataTypes.STRING()).build()
        table = self.t_env.from_descriptor(
            TableDescriptor.for_connector("datagen")
            .option("number-of-rows", "10")
            .schema(schema)
            .build()
        )
        self.t_env.create_temporary_table(
            "RegisteredSink",
            TableDescriptor.for_connector("blackhole").schema(schema).build(),
        )
        table_pipeline = table.insert_into("RegisteredSink")
        object_identifier = table_pipeline.get_sink_identifier()

        self.assertEqual(object_identifier.get_catalog_name(), "default_catalog")
        self.assertEqual(object_identifier.get_database_name(), "default_database")
        self.assertEqual(object_identifier.get_object_name(), "RegisteredSink")
        self.assertEqual(
            object_identifier.as_summary_string(),
            "default_catalog.default_database.RegisteredSink",
        )

    def test_table_pipeline_execute(self):
        schema = Schema.new_builder().column("f0", DataTypes.STRING()).build()
        table = self.t_env.from_descriptor(
            TableDescriptor.for_connector("datagen")
            .option("number-of-rows", "10")
            .schema(schema)
            .build()
        )
        self.t_env.create_temporary_table(
            "RegisteredSinkForExecute",
            TableDescriptor.for_connector("blackhole").schema(schema).build(),
        )
        table_pipeline = table.insert_into("RegisteredSinkForExecute")

        table_result = table_pipeline.execute()
        self.assertEqual(
            table_result.get_result_kind(), ResultKind.SUCCESS_WITH_CONTENT
        )
        self.assertEqual(
            table_result.get_table_schema().get_field_names()[0],
            "default_catalog.default_database.RegisteredSinkForExecute",
        )

    def test_table_pipeline_compile(self):
        schema = Schema.new_builder().column("f0", DataTypes.STRING()).build()
        table = self.t_env.from_descriptor(
            TableDescriptor.for_connector("datagen")
            .option("number-of-rows", "10")
            .schema(schema)
            .build()
        )
        self.t_env.create_temporary_table(
            "RegisteredSinkForPlanCompile",
            TableDescriptor.for_connector("blackhole").schema(schema).build(),
        )
        table_pipeline = table.insert_into("RegisteredSinkForPlanCompile")
        compiled_plan = table_pipeline.compile_plan()
        self.assertIsNotNone(compiled_plan.as_json_string())


if __name__ == "__main__":
    import unittest

    try:
        import xmlrunner

        testRunner = xmlrunner.XMLTestRunner(output="target/test-reports")
    except ImportError:
        testRunner = None
    unittest.main(testRunner=testRunner, verbosity=2)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import os
import shutil
import socket
import subprocess
import sys
import tempfile
import time
import unittest

import grpc
from concurrent import futures
from apache_beam.portability.api.org.apache.beam.model.fn_execution.v1.beam_provision_api_pb2 \
    import ProvisionInfo, GetProvisionInfoResponse
from apache_beam.portability.api.org.apache.beam.model.fn_execution.v1.beam_provision_api_pb2_grpc \
    import ProvisionServiceServicer, add_ProvisionServiceServicer_to_server
from google.protobuf import json_format

from pyflink.java_gateway import get_gateway
from pyflink.pyflink_gateway_server import on_windows
from pyflink.testing.test_case_utils import PyFlinkTestCase


class PythonBootTests(PyFlinkTestCase):

    def setUp(self):
        provision_info = json_format.Parse('{"retrievalToken": "test_token"}', ProvisionInfo())
        response = GetProvisionInfoResponse(info=provision_info)

        def get_unused_port():
            sock = socket.socket()
            sock.bind(('', 0))
            port = sock.getsockname()[1]
            sock.close()
            return port

        class ProvisionService(ProvisionServiceServicer):
            def GetProvisionInfo(self, request, context):
                return response

        def start_test_provision_server():
            server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
            add_ProvisionServiceServicer_to_server(ProvisionService(), server)
            port = get_unused_port()
            server.add_insecure_port('[::]:' + str(port))
            server.start()
            return server, port

        self.provision_server, self.provision_port = start_test_provision_server()

        self.env = dict(os.environ)
        self.env["python"] = sys.executable
        self.env["FLINK_BOOT_TESTING"] = "1"
        self.env["BOOT_LOG_DIR"] = os.path.join(self.env["FLINK_HOME"], "log")

        self.tmp_dir = tempfile.mkdtemp(str(time.time()), dir=self.tempdir)
        # assume that this file is in flink-python source code directory.
        pyflink_package_dir = os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        runner_script = "pyflink-udf-runner.bat" if on_windows() else \
            "pyflink-udf-runner.sh"
        self.runner_path = os.path.join(
            pyflink_package_dir, "bin", runner_script)

    def run_boot_py(self):
        args = [self.runner_path, "--id", "1",
                "--logging_endpoint", "localhost:0000",
                "--artifact_endpoint", "whatever",
                "--provision_endpoint", "localhost:%d" % self.provision_port,
                "--control_endpoint", "localhost:0000",
                "--semi_persist_dir", self.tmp_dir]

        return subprocess.call(args, env=self.env)

    def test_python_boot(self):
        exit_code = self.run_boot_py()
        self.assertTrue(exit_code == 0, "the boot.py exited with non-zero code.")

    @unittest.skipIf(on_windows(), "'subprocess.check_output' in Windows always return empty "
                                   "string, skip this test.")
    def test_param_validation(self):
        args = [self.runner_path]
        exit_message = subprocess.check_output(args, env=self.env).decode("utf-8")
        self.assertIn("No id provided.", exit_message)

        args = [self.runner_path, "--id", "1"]
        exit_message = subprocess.check_output(args, env=self.env).decode("utf-8")
        self.assertIn("No provision endpoint provided.", exit_message)

    def test_set_working_directory(self):
        JProcessPythonEnvironmentManager = \
            get_gateway().jvm.org.apache.flink.python.env.process.ProcessPythonEnvironmentManager

        output_file = os.path.join(self.tmp_dir, "output.txt")
        pyflink_dir = os.path.join(self.tmp_dir, "pyflink")
        os.mkdir(pyflink_dir)
        # just create an empty file
        open(os.path.join(pyflink_dir, "__init__.py"), 'a').close()
        fn_execution_dir = os.path.join(pyflink_dir, "fn_execution")
        os.mkdir(fn_execution_dir)
        open(os.path.join(fn_execution_dir, "__init__.py"), 'a').close()
        beam_dir = os.path.join(fn_execution_dir, "beam")
        os.mkdir(beam_dir)
        open(os.path.join(beam_dir, "__init__.py"), 'a').close()
        with open(os.path.join(beam_dir, "beam_boot.py"), "w") as f:
            f.write("import os\nwith open(r'%s', 'w') as f:\n    f.write(os.getcwd())" %
                    output_file)

        # test if the name of working directory variable of udf runner is consist with
        # ProcessPythonEnvironmentManager.
        self.env[JProcessPythonEnvironmentManager.PYTHON_WORKING_DIR] = self.tmp_dir
        self.env["python"] = sys.executable
        args = [self.runner_path]
        subprocess.check_output(args, env=self.env)
        process_cwd = None
        if os.path.exists(output_file):
            with open(output_file, 'r') as f:
                process_cwd = f.read()

        self.assertEqual(os.path.realpath(self.tmp_dir),
                         process_cwd,
                         "setting working directory variable is not work!")

    def tearDown(self):
        self.provision_server.stop(0)
        try:
            if self.tmp_dir is not None:
                shutil.rmtree(self.tmp_dir)
        except:
            pass



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.common.config_options import ConfigOptions
from pyflink.table import DataTypes
from pyflink.table.schema import Schema
from pyflink.table.table_descriptor import TableDescriptor, FormatDescriptor
from pyflink.testing.test_case_utils import PyFlinkStreamTableTestCase


class TableDescriptorTest(PyFlinkStreamTableTestCase):

    def setUp(self):
        super(TableDescriptorTest, self).setUp()
        self.option_a = ConfigOptions.key("a").boolean_type().no_default_value()
        self.option_b = ConfigOptions.key("b").int_type().no_default_value()
        self.key_format = ConfigOptions.key("key.format").string_type().no_default_value()

    def test_basic(self):
        schema = Schema.new_builder() \
            .column("f0", DataTypes.STRING()) \
            .column("f1", DataTypes.BIGINT()) \
            .primary_key("f0") \
            .build()

        descriptor = TableDescriptor.for_connector("test-connector") \
            .schema(schema) \
            .partitioned_by("f0") \
            .comment("Test Comment") \
            .build()

        self.assertIsNotNone(descriptor.get_schema())

        self.assertEqual(1, len(descriptor.get_partition_keys()))
        self.assertEqual("f0", descriptor.get_partition_keys()[0])

        self.assertEqual(1, len(descriptor.get_options()))
        self.assertEqual("test-connector", descriptor.get_options().get("connector"))

        self.assertEqual("Test Comment", descriptor.get_comment())

    def test_no_schema(self):
        descriptor = TableDescriptor.for_connector("test-connector").build()
        self.assertIsNone(descriptor.get_schema())

    def test_options(self):
        descriptor = TableDescriptor.for_connector("test-connector") \
            .schema(Schema.new_builder().build()) \
            .option(self.option_a, False) \
            .option(self.option_b, 42) \
            .option("c", "C") \
            .build()
        self.assertEqual(4, len(descriptor.get_options()))
        self.assertEqual("test-connector", descriptor.get_options().get("connector"))
        self.assertEqual("false", descriptor.get_options().get("a"))
        self.assertEqual("42", descriptor.get_options().get("b"))
        self.assertEqual("C", descriptor.get_options().get("c"))

    def test_format_basic(self):
        descriptor = TableDescriptor.for_connector("test-connector") \
            .schema(Schema.new_builder().build()) \
            .format("json") \
            .build()
        self.assertEqual(2, len(descriptor.get_options()))
        self.assertEqual("test-connector", descriptor.get_options().get("connector"))
        self.assertEqual("json", descriptor.get_options().get("format"))

    def test_format_with_format_descriptor(self):
        descriptor = TableDescriptor.for_connector("test-connector") \
            .schema(Schema.new_builder().build()) \
            .format(FormatDescriptor.for_format("test-format")
                    .option(self.option_a, True)
                    .option(self.option_b, 42)
                    .option("c", "C")
                    .build(),
                    self.key_format) \
            .build()
        self.assertEqual(5, len(descriptor.get_options()))
        self.assertEqual("test-connector", descriptor.get_options().get("connector"))
        self.assertEqual("test-format", descriptor.get_options().get("key.format"))
        self.assertEqual("true", descriptor.get_options().get("key.test-format.a"))
        self.assertEqual("42", descriptor.get_options().get("key.test-format.b"))
        self.assertEqual("C", descriptor.get_options().get("key.test-format.c"))

    def test_to_string(self):
        schema = Schema.new_builder().column("f0", DataTypes.STRING()).build()
        format_descriptor = FormatDescriptor \
            .for_format("test-format") \
            .option(self.option_a, False) \
            .build()
        table_descriptor = TableDescriptor.for_connector("test-connector") \
            .schema(schema) \
            .partitioned_by("f0") \
            .option(self.option_a, True) \
            .format(format_descriptor) \
            .comment("Test Comment") \
            .build()
        self.assertEqual("test-format[{a=false}]", str(format_descriptor))
        self.assertEqual("""(
  `f0` STRING
)
COMMENT 'Test Comment'
PARTITIONED BY (`f0`)
WITH (
  'a' = 'true',
  'connector' = 'test-connector',
  'test-format.a' = 'false',
  'format' = 'test-format'
)""", str(table_descriptor))

    def test_execute_insert_to_table_descriptor(self):
        schema = Schema.new_builder() \
            .column("f0", DataTypes.STRING()) \
            .build()
        table = self.t_env.from_descriptor(TableDescriptor
                                           .for_connector("datagen")
                                           .option("number-of-rows", '10')
                                           .schema(schema)
                                           .build())
        table_result = table.execute_insert(TableDescriptor
                                            .for_connector("blackhole")
                                            .schema(schema)
                                            .build())
        table_result.collect()

    def test_statement_set_insert_using_table_descriptor(self):
        schema = Schema.new_builder() \
            .column("f0", DataTypes.INT()) \
            .build()

        source_descriptor = TableDescriptor.for_connector("datagen") \
            .schema(schema) \
            .option("number-of-rows", '10') \
            .build()

        sink_descriptor = TableDescriptor.for_connector("blackhole") \
            .schema(schema) \
            .build()

        self.t_env.create_temporary_table("T", source_descriptor)

        stmt_set = self.t_env.create_statement_set()
        stmt_set.add_insert(sink_descriptor, self.t_env.from_path("T"))

        stmt_set.execute().wait()


if __name__ == '__main__':
    import unittest

    try:
        import xmlrunner

        testRunner = xmlrunner.XMLTestRunner(output='target/test-reports')
    except ImportError:
        testRunner = None
    unittest.main(testRunner=testRunner, verbosity=2)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import unittest

from pyflink.common import Types
from pyflink.datastream.connectors.cassandra import CassandraSink, MapperOptions, ConsistencyLevel
from pyflink.testing.test_case_utils import PyFlinkStreamingTestCase


class CassandraSinkTest(PyFlinkStreamingTestCase):

    @unittest.skip("Disable due to cassandra connectors is not support 2.0 for new.")
    def test_cassandra_sink(self):
        type_info = Types.ROW([Types.STRING(), Types.INT()])
        ds = self.env.from_collection([('ab', 1), ('bdc', 2), ('cfgs', 3), ('deeefg', 4)],
                                      type_info=type_info)
        cassandra_sink_builder = CassandraSink.add_sink(ds)

        cassandra_sink = cassandra_sink_builder\
            .set_host('localhost', 9876) \
            .set_query('query') \
            .enable_ignore_null_fields() \
            .set_mapper_options(MapperOptions()
                                .ttl(1)
                                .timestamp(100)
                                .tracing(True)
                                .if_not_exists(False)
                                .consistency_level(ConsistencyLevel.ANY)
                                .save_null_fields(True)) \
            .set_max_concurrent_requests(1000) \
            .build()

        cassandra_sink.name('cassandra_sink').set_parallelism(3)

        plan = eval(self.env.get_execution_plan())
        self.assertEqual("Sink: cassandra_sink", plan['nodes'][1]['type'])
        self.assertEqual(3, plan['nodes'][1]['parallelism'])



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.datastream.state_backend import (_from_j_state_backend, CustomStateBackend,
                                              PredefinedOptions, EmbeddedRocksDBStateBackend)
from pyflink.java_gateway import get_gateway
from pyflink.pyflink_gateway_server import on_windows
from pyflink.testing.test_case_utils import PyFlinkTestCase
from pyflink.util.java_utils import load_java_class


class EmbeddedRocksDBStateBackendTests(PyFlinkTestCase):

    def test_create_rocks_db_state_backend(self):

        self.assertIsNotNone(EmbeddedRocksDBStateBackend())

        self.assertIsNotNone(EmbeddedRocksDBStateBackend(True))

        self.assertIsNotNone(EmbeddedRocksDBStateBackend(False))

    def test_get_set_db_storage_paths(self):
        if on_windows():
            storage_path = ["file:/C:/var/db_storage_dir1/",
                            "file:/C:/var/db_storage_dir2/",
                            "file:/C:/var/db_storage_dir3/"]
            expected = ["C:\\var\\db_storage_dir1",
                        "C:\\var\\db_storage_dir2",
                        "C:\\var\\db_storage_dir3"]
        else:
            storage_path = ["file://var/db_storage_dir1/",
                            "file://var/db_storage_dir2/",
                            "file://var/db_storage_dir3/"]
            expected = ["/db_storage_dir1",
                        "/db_storage_dir2",
                        "/db_storage_dir3"]

        state_backend = EmbeddedRocksDBStateBackend()
        state_backend.set_db_storage_paths(*storage_path)
        self.assertEqual(state_backend.get_db_storage_paths(), expected)

    def test_get_set_predefined_options(self):

        state_backend = EmbeddedRocksDBStateBackend()

        self.assertEqual(state_backend.get_predefined_options(), PredefinedOptions.DEFAULT)

        state_backend.set_predefined_options(PredefinedOptions.SPINNING_DISK_OPTIMIZED_HIGH_MEM)

        self.assertEqual(state_backend.get_predefined_options(),
                         PredefinedOptions.SPINNING_DISK_OPTIMIZED_HIGH_MEM)

        state_backend.set_predefined_options(PredefinedOptions.SPINNING_DISK_OPTIMIZED)

        self.assertEqual(state_backend.get_predefined_options(),
                         PredefinedOptions.SPINNING_DISK_OPTIMIZED)

        state_backend.set_predefined_options(PredefinedOptions.FLASH_SSD_OPTIMIZED)

        self.assertEqual(state_backend.get_predefined_options(),
                         PredefinedOptions.FLASH_SSD_OPTIMIZED)

        state_backend.set_predefined_options(PredefinedOptions.DEFAULT)

        self.assertEqual(state_backend.get_predefined_options(), PredefinedOptions.DEFAULT)

    def test_get_set_options(self):

        state_backend = EmbeddedRocksDBStateBackend()

        self.assertIsNone(state_backend.get_options())

        state_backend.set_options(
            "org.apache.flink.state.rocksdb."
            "RocksDBStateBackendConfigTest$TestOptionsFactory")

        self.assertEqual(state_backend.get_options(),
                         "org.apache.flink.state.rocksdb."
                         "RocksDBStateBackendConfigTest$TestOptionsFactory")

    def test_get_set_number_of_transfer_threads(self):

        state_backend = EmbeddedRocksDBStateBackend()

        self.assertEqual(state_backend.get_number_of_transfer_threads(), 4)

        state_backend.set_number_of_transfer_threads(8)

        self.assertEqual(state_backend.get_number_of_transfer_threads(), 8)


class CustomStateBackendTests(PyFlinkTestCase):

    def test_create_custom_state_backend(self):
        gateway = get_gateway()
        JConfiguration = gateway.jvm.org.apache.flink.configuration.Configuration
        j_config = JConfiguration()
        j_factory = load_java_class("org.apache.flink.streaming.runtime.tasks."
                                    "StreamTaskTest$TestMemoryStateBackendFactory").newInstance()
        context_classloader = gateway.jvm.Thread.currentThread().getContextClassLoader()
        state_backend = _from_j_state_backend(j_factory.createFromConfig(j_config,
                                                                         context_classloader))

        self.assertIsInstance(state_backend, CustomStateBackend)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from typing import List

from pyflink.table.expression import Expression
from pyflink.table.types import DataType, _from_java_data_type
from pyflink.util.api_stability_decorators import PublicEvolving

__all__ = ["ResolvedExpression"]


@PublicEvolving()
class ResolvedExpression(Expression):
    """
    Expression that has been fully resolved and validated.

    Compared to :class:`~pyflink.table.Expression`, resolved expressions do not contain unresolved
    subexpressions anymore and provide an output data type for the computation result.

    Instances of this class describe a fully parameterized, immutable expression that can be
    serialized and persisted.

    Resolved expression are the output of the API to the planner and are pushed from the planner
    into interfaces, for example, for predicate push-down.
    """

    def __init__(self, j_resolved_expr):
        super().__init__(j_expr_or_property_name=j_resolved_expr)
        self._j_resolved_expr = j_resolved_expr

    def __str__(self):
        return self._j_resolved_expr.toString()

    def get_output_data_type(self) -> DataType:
        """
        Returns the data type of the computation result.
        """
        j_data_type = self._j_resolved_expr.getOutputDataType()
        return _from_java_data_type(j_data_type)

    def get_resolved_children(self) -> List["ResolvedExpression"]:
        j_resolved_children = self._j_resolved_expr.getResolvedChildren()
        return [ResolvedExpression(j_resolved_child) for j_resolved_child in j_resolved_children]

    def as_serializable_string(self) -> str:
        """
        Returns a string that fully serializes this instance. The serialized string can be used for
        storing the query in, for example, a :class:`~pyflink.table.catalog.Catalog` as a view.
        """
        return self._j_resolved_expr.asSerializableString()

    def as_summary_string(self) -> str:
        """
        Returns a string that summarizes this expression for printing to a console. An
        implementation might skip very specific properties.
        """
        return self._j_resolved_expr.asSummaryString()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.common import Types
from pyflink.datastream.formats.json import JsonRowSerializationSchema, JsonRowDeserializationSchema
from pyflink.java_gateway import get_gateway
from pyflink.testing.test_case_utils import PyFlinkTestCase


class JsonSerializationSchemaTests(PyFlinkTestCase):

    def test_json_row_serialization_deserialization_schema(self):
        jvm = get_gateway().jvm
        jsons = ["{\"svt\":\"2020-02-24T12:58:09.209+0800\"}",
                 "{\"svt\":\"2020-02-24T12:58:09.209+0800\", "
                 "\"ops\":{\"id\":\"281708d0-4092-4c21-9233-931950b6eccf\"},\"ids\":[1, 2, 3]}",
                 "{\"svt\":\"2020-02-24T12:58:09.209+0800\"}"]
        expected_jsons = ["{\"svt\":\"2020-02-24T12:58:09.209+0800\",\"ops\":null,\"ids\":null}",
                          "{\"svt\":\"2020-02-24T12:58:09.209+0800\","
                          "\"ops\":{\"id\":\"281708d0-4092-4c21-9233-931950b6eccf\"},"
                          "\"ids\":[1,2,3]}",
                          "{\"svt\":\"2020-02-24T12:58:09.209+0800\",\"ops\":null,\"ids\":null}"]

        row_schema = Types.ROW_NAMED(["svt", "ops", "ids"],
                                     [Types.STRING(),
                                     Types.ROW_NAMED(['id'], [Types.STRING()]),
                                     Types.PRIMITIVE_ARRAY(Types.INT())])

        json_row_serialization_schema = JsonRowSerializationSchema.builder() \
            .with_type_info(row_schema).build()
        json_row_deserialization_schema = JsonRowDeserializationSchema.builder() \
            .type_info(row_schema).build()
        json_row_serialization_schema._j_serialization_schema.open(
            jvm.org.apache.flink.connector.testutils.formats.DummyInitializationContext())
        json_row_deserialization_schema._j_deserialization_schema.open(
            jvm.org.apache.flink.connector.testutils.formats.DummyInitializationContext())

        for i in range(len(jsons)):
            j_row = json_row_deserialization_schema._j_deserialization_schema\
                .deserialize(bytes(jsons[i], encoding='utf-8'))
            result = str(json_row_serialization_schema._j_serialization_schema.serialize(j_row),
                         encoding='utf-8')
            self.assertEqual(expected_jsons[i], result)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.datastream.connectors.number_seq import NumberSequenceSource
from pyflink.testing.test_case_utils import PyFlinkStreamingTestCase
from pyflink.util.java_utils import load_java_class


class SequenceSourceTests(PyFlinkStreamingTestCase):

    def test_seq_source(self):
        seq_source = NumberSequenceSource(1, 10)

        seq_source_clz = load_java_class(
            "org.apache.flink.api.connector.source.lib.NumberSequenceSource")
        from_field = seq_source_clz.getDeclaredField("from")
        from_field.setAccessible(True)
        self.assertEqual(1, from_field.get(seq_source.get_java_function()))

        to_field = seq_source_clz.getDeclaredField("to")
        to_field.setAccessible(True)
        self.assertEqual(10, to_field.get(seq_source.get_java_function()))



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import datetime
import decimal

from pandas.testing import assert_frame_equal

from pyflink.common import Row
from pyflink.table.types import DataTypes
from pyflink.testing import source_sink_utils
from pyflink.testing.test_case_utils import PyFlinkBatchTableTestCase, \
    PyFlinkStreamTableTestCase


class PandasConversionTestBase(object):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.data = [(1, 1, 1, 1, True, 1.1, 1.2, 'hello', bytearray(b"aaa"),
                     decimal.Decimal('1000000000000000000.01'), datetime.date(2014, 9, 13),
                     datetime.time(hour=1, minute=0, second=1),
                     datetime.datetime(1970, 1, 1, 0, 0, 0, 123000), ['hello', '中文'],
                     Row(a=1, b='hello', c=datetime.datetime(1970, 1, 1, 0, 0, 0, 123000),
                         d=[1, 2])),
                    (1, 2, 2, 2, False, 2.1, 2.2, 'world', bytearray(b"bbb"),
                     decimal.Decimal('1000000000000000000.02'), datetime.date(2014, 9, 13),
                     datetime.time(hour=1, minute=0, second=1),
                     datetime.datetime(1970, 1, 1, 0, 0, 0, 123000), ['hello', '中文'],
                     Row(a=1, b='hello', c=datetime.datetime(1970, 1, 1, 0, 0, 0, 123000),
                         d=[1, 2]))]
        cls.data_type = DataTypes.ROW(
            [DataTypes.FIELD("f1", DataTypes.TINYINT()),
             DataTypes.FIELD("f2", DataTypes.SMALLINT()),
             DataTypes.FIELD("f3", DataTypes.INT()),
             DataTypes.FIELD("f4", DataTypes.BIGINT()),
             DataTypes.FIELD("f5", DataTypes.BOOLEAN()),
             DataTypes.FIELD("f6", DataTypes.FLOAT()),
             DataTypes.FIELD("f7", DataTypes.DOUBLE()),
             DataTypes.FIELD("f8", DataTypes.STRING()),
             DataTypes.FIELD("f9", DataTypes.BYTES()),
             DataTypes.FIELD("f10", DataTypes.DECIMAL(38, 18)),
             DataTypes.FIELD("f11", DataTypes.DATE()),
             DataTypes.FIELD("f12", DataTypes.TIME()),
             DataTypes.FIELD("f13", DataTypes.TIMESTAMP(3)),
             DataTypes.FIELD("f14", DataTypes.ARRAY(DataTypes.STRING())),
             DataTypes.FIELD("f15", DataTypes.ROW(
                 [DataTypes.FIELD("a", DataTypes.INT()),
                  DataTypes.FIELD("b", DataTypes.STRING()),
                  DataTypes.FIELD("c", DataTypes.TIMESTAMP(3)),
                  DataTypes.FIELD("d", DataTypes.ARRAY(DataTypes.INT()))]))], False)
        cls.pdf = cls.create_pandas_data_frame()

    @classmethod
    def create_pandas_data_frame(cls):
        data_dict = {}
        for j, name in enumerate(cls.data_type.names):
            data_dict[name] = [cls.data[i][j] for i in range(len(cls.data))]
        # need convert to numpy types
        import numpy as np
        data_dict["f1"] = np.int8(data_dict["f1"])
        data_dict["f2"] = np.int16(data_dict["f2"])
        data_dict["f3"] = np.int32(data_dict["f3"])
        data_dict["f4"] = np.int64(data_dict["f4"])
        data_dict["f6"] = np.float32(data_dict["f6"])
        data_dict["f7"] = np.float64(data_dict["f7"])
        data_dict["f15"] = [row.as_dict() for row in data_dict["f15"]]
        import pandas as pd
        return pd.DataFrame(data=data_dict,
                            index=[2., 3.],
                            columns=['f1', 'f2', 'f3', 'f4', 'f5', 'f6', 'f7', 'f8', 'f9',
                                     'f10', 'f11', 'f12', 'f13', 'f14', 'f15'])


class PandasConversionTests(PandasConversionTestBase):

    def test_from_pandas_with_incorrect_schema(self):
        fields = self.data_type.fields.copy()
        fields[0], fields[7] = fields[7], fields[0]  # swap str with tinyint
        wrong_schema = DataTypes.ROW(fields)  # should be DataTypes.STRING()
        with self.assertRaisesRegex(Exception, "Expected a string.*got int8"):
            self.t_env.from_pandas(self.pdf, schema=wrong_schema)

    def test_from_pandas_with_names(self):
        # skip decimal as currently only decimal(38, 18) is supported
        pdf = self.pdf.drop(['f10', 'f11', 'f12', 'f13', 'f14', 'f15'], axis=1)
        new_names = list(map(str, range(len(pdf.columns))))
        table = self.t_env.from_pandas(pdf, schema=new_names)
        self.assertEqual(new_names, table.get_schema().get_field_names())
        table = self.t_env.from_pandas(pdf, schema=tuple(new_names))
        self.assertEqual(new_names, table.get_schema().get_field_names())

    def test_from_pandas_with_types(self):
        new_types = self.data_type.field_types()
        new_types[0] = DataTypes.BIGINT()
        table = self.t_env.from_pandas(self.pdf, schema=new_types)
        self.assertEqual(new_types, table.get_schema().get_field_data_types())
        table = self.t_env.from_pandas(self.pdf, schema=tuple(new_types))
        self.assertEqual(new_types, table.get_schema().get_field_data_types())


class PandasConversionITTests(PandasConversionTestBase):

    def test_from_pandas(self):
        table = self.t_env.from_pandas(self.pdf, self.data_type, 5)
        self.assertEqual(self.data_type, table.get_schema().to_row_data_type())

        table = table.filter(table.f2 < 2)
        sink_table_ddl = """
            CREATE TABLE Results(
            f1 TINYINT,
            f2 SMALLINT,
            f3 INT,
            f4 BIGINT,
            f5 BOOLEAN,
            f6 FLOAT,
            f7 DOUBLE,
            f8 STRING,
            f9 BYTES,
            f10 DECIMAL(38, 18),
            f11 DATE,
            f12 TIME,
            f13 TIMESTAMP(3),
            f14 ARRAY<STRING>,
            f15 ROW<a INT, b STRING, c TIMESTAMP(3), d ARRAY<INT>>)
            WITH ('connector'='test-sink')
        """
        self.t_env.execute_sql(sink_table_ddl)

        table.execute_insert("Results").wait()
        actual = source_sink_utils.results()
        self.assert_equals(actual,
                           ["+I[1, 1, 1, 1, true, 1.1, 1.2, hello, [97, 97, 97], "
                            "1000000000000000000.010000000000000000, 2014-09-13, 01:00:01, "
                            "1970-01-01T00:00:00.123, [hello, 中文], +I[1, hello, "
                            "1970-01-01T00:00:00.123, [1, 2]]]"])

    def test_to_pandas(self):
        table = self.t_env.from_pandas(self.pdf, self.data_type)
        result_pdf = table.to_pandas()
        result_pdf.index = self.pdf.index
        self.assertEqual(2, len(result_pdf))
        expected_arrow = self.pdf.to_records(index=False)
        result_arrow = result_pdf.to_records(index=False)
        for r in range(len(expected_arrow)):
            for e in range(len(expected_arrow[r])):
                self.assert_equal_field(expected_arrow[r][e], result_arrow[r][e])

    def test_empty_to_pandas(self):
        table = self.t_env.from_pandas(self.pdf, self.data_type)
        pdf = table.filter(table.f1 < 0).to_pandas()
        self.assertTrue(pdf.empty)

    def test_to_pandas_for_retract_table(self):
        table = self.t_env.from_pandas(self.pdf, self.data_type)
        result_pdf = table.group_by(table.f1).select(table.f2.max.alias('f2')).to_pandas()
        import pandas as pd
        import numpy as np
        assert_frame_equal(result_pdf, pd.DataFrame(data={'f2': np.int16([2])}))

        result_pdf = table.group_by(table.f2).select(table.f1.max.alias('f2')).to_pandas()
        assert_frame_equal(result_pdf, pd.DataFrame(data={'f2': np.int8([1, 1])}))

    def assert_equal_field(self, expected_field, result_field):
        import numpy as np
        result_type = type(result_field)
        if result_type == dict:
            self.assertEqual(expected_field.keys(), result_field.keys())
            for key in expected_field:
                self.assert_equal_field(expected_field[key], result_field[key])
        elif result_type == np.ndarray:
            self.assertTrue((expected_field == result_field).all())
        else:
            self.assertTrue(expected_field == result_field)


class BatchPandasConversionTests(PandasConversionTests,
                                 PandasConversionITTests,
                                 PyFlinkBatchTableTestCase):
    pass


class StreamPandasConversionTests(PandasConversionITTests,
                                  PyFlinkStreamTableTestCase):
    def test_to_pandas_with_event_time(self):
        self.t_env.get_config().set("parallelism.default", "1")
        # create source file path
        import tempfile
        import os
        tmp_dir = tempfile.gettempdir()
        data = [
            '2018-03-11 03:10:00',
            '2018-03-11 03:10:00',
            '2018-03-11 03:10:00',
            '2018-03-11 03:40:00',
            '2018-03-11 04:20:00',
            '2018-03-11 03:30:00'
        ]
        source_path = tmp_dir + '/test_to_pandas_with_event_time.csv'
        with open(source_path, 'w') as fd:
            for ele in data:
                fd.write(ele + '\n')

        self.t_env.get_config().set(
            "pipeline.time-characteristic", "EventTime")

        source_table = """
            create table source_table(
                rowtime TIMESTAMP(3),
                WATERMARK FOR rowtime AS rowtime - INTERVAL '60' MINUTE
            ) with(
                'connector.type' = 'filesystem',
                'format.type' = 'csv',
                'connector.path' = '%s',
                'format.ignore-first-line' = 'false',
                'format.field-delimiter' = ','
            )
        """ % source_path
        self.t_env.execute_sql(source_table)
        t = self.t_env.from_path("source_table")
        result_pdf = t.to_pandas()
        import pandas as pd
        os.remove(source_path)
        expected_df = pd.DataFrame(data={"rowtime": pd.Series(data, dtype="datetime64[ms]")})
        assert_frame_equal(result_pdf, expected_df)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.java_gateway import get_gateway

from pyflink.common import Configuration
from pyflink.table import EnvironmentSettings
from pyflink.testing.test_case_utils import PyFlinkTestCase


class EnvironmentSettingsTests(PyFlinkTestCase):

    def test_mode_selection(self):

        builder = EnvironmentSettings.new_instance()

        # test the default behaviour to make sure it is consistent with the python doc
        environment_settings = builder.build()
        self.assertTrue(environment_settings.is_streaming_mode())

        # test in_streaming_mode
        environment_settings = builder.in_streaming_mode().build()
        self.assertTrue(environment_settings.is_streaming_mode())

        environment_settings = EnvironmentSettings.in_streaming_mode()
        self.assertTrue(environment_settings.is_streaming_mode())

        # test in_batch_mode
        environment_settings = builder.in_batch_mode().build()
        self.assertFalse(environment_settings.is_streaming_mode())

        environment_settings = EnvironmentSettings.in_batch_mode()
        self.assertFalse(environment_settings.is_streaming_mode())

    def test_with_built_in_catalog_name(self):

        gateway = get_gateway()

        DEFAULT_BUILTIN_CATALOG = gateway.jvm.TableConfigOptions.TABLE_CATALOG_NAME.defaultValue()

        builder = EnvironmentSettings.new_instance()

        # test the default behaviour to make sure it is consistent with the python doc
        environment_settings = builder.build()

        self.assertEqual(environment_settings.get_built_in_catalog_name(), DEFAULT_BUILTIN_CATALOG)

        environment_settings = builder.with_built_in_catalog_name("my_catalog").build()

        self.assertEqual(environment_settings.get_built_in_catalog_name(), "my_catalog")

    def test_with_built_in_database_name(self):

        gateway = get_gateway()

        DEFAULT_BUILTIN_DATABASE = gateway.jvm.TableConfigOptions.TABLE_DATABASE_NAME.defaultValue()

        builder = EnvironmentSettings.new_instance()

        # test the default behaviour to make sure it is consistent with the python doc
        environment_settings = builder.build()

        self.assertEqual(environment_settings.get_built_in_database_name(),
                         DEFAULT_BUILTIN_DATABASE)

        environment_settings = builder.with_built_in_database_name("my_database").build()

        self.assertEqual(environment_settings.get_built_in_database_name(), "my_database")

    def test_to_configuration(self):

        expected_settings = EnvironmentSettings.new_instance().in_batch_mode().build()
        config = expected_settings.get_configuration()

        self.assertEqual("BATCH", config.get_string("execution.runtime-mode", "stream"))

    def test_from_configuration(self):

        config = Configuration()
        config.set_string("execution.runtime-mode", "batch")

        actual_setting = EnvironmentSettings.new_instance().with_configuration(config).build()
        self.assertFalse(actual_setting.is_streaming_mode(), "Use batch mode.")



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import inspect

from pyflink.common import Time, TypeInformation
from pyflink.datastream import async_retry_strategies
from pyflink.datastream.data_stream import DataStream, _get_one_input_stream_operator
from pyflink.datastream.functions import AsyncFunctionDescriptor, AsyncFunction, AsyncRetryStrategy
from pyflink.java_gateway import get_gateway
from pyflink.util.java_utils import get_j_env_configuration


class AsyncDataStream(object):
    """
    A helper class to apply :class:`~AsyncFunction` to a data stream.
    """

    @staticmethod
    def unordered_wait(
            data_stream: DataStream,
            async_function: AsyncFunction,
            timeout: Time,
            capacity: int = 100,
            output_type: TypeInformation = None) -> 'DataStream':
        """
        Adds an async function to the data stream. The order of output stream records may be
        reordered.

        :param data_stream: The input data stream.
        :param async_function: The async function.
        :param timeout: The timeout for the asynchronous operation to complete.
        :param capacity: The max number of async i/o operation that can be triggered.
        :param output_type: The output data type.
        :return: The transformed DataStream.
        """
        return AsyncDataStream.unordered_wait_with_retry(
            data_stream, async_function, timeout, async_retry_strategies.NO_RETRY_STRATEGY,
            capacity, output_type)

    @staticmethod
    def unordered_wait_with_retry(
            data_stream: DataStream,
            async_function: AsyncFunction,
            timeout: Time,
            async_retry_strategy: AsyncRetryStrategy,
            capacity: int = 100,
            output_type: TypeInformation = None) -> 'DataStream':
        """
        Adds an async function with an AsyncRetryStrategy to support retry of AsyncFunction to the
        data stream. The order of output stream records may be reordered.

        :param data_stream: The input data stream.
        :param async_function: The async function.
        :param timeout: The timeout for the asynchronous operation to complete.
        :param async_retry_strategy: The strategy of reattempt async i/o operation that can be
                                     triggered
        :param capacity: The max number of async i/o operation that can be triggered.
        :param output_type: The output data type.
        :return: The transformed DataStream.
        """
        AsyncDataStream._validate(data_stream, async_function, timeout, async_retry_strategy)

        from pyflink.fn_execution import flink_fn_execution_pb2
        j_python_data_stream_function_operator, j_output_type_info = \
            _get_one_input_stream_operator(
                data_stream,
                AsyncFunctionDescriptor(
                    async_function, timeout, capacity, async_retry_strategy,
                    AsyncFunctionDescriptor.OutputMode.UNORDERED),
                flink_fn_execution_pb2.UserDefinedDataStreamFunction.PROCESS,  # type: ignore
                output_type)
        return DataStream(data_stream._j_data_stream.transform(
            "async wait operator",
            j_output_type_info,
            j_python_data_stream_function_operator))

    @staticmethod
    def ordered_wait(
            data_stream: DataStream,
            async_function: AsyncFunction,
            timeout: Time,
            capacity: int = 100,
            output_type: TypeInformation = None) -> 'DataStream':
        """
        Adds an async function to the data stream. The order to process input records
        is guaranteed to be the same as input ones.

        :param data_stream: The input data stream.
        :param async_function: The async function.
        :param timeout: The timeout for the asynchronous operation to complete.
        :param capacity: The max number of async i/o operation that can be triggered.
        :param output_type: The output data type.
        :return: The transformed DataStream.
        """
        return AsyncDataStream.ordered_wait_with_retry(
            data_stream, async_function, timeout, async_retry_strategies.NO_RETRY_STRATEGY,
            capacity, output_type)

    @staticmethod
    def ordered_wait_with_retry(
            data_stream: DataStream,
            async_function: AsyncFunction,
            timeout: Time,
            async_retry_strategy: AsyncRetryStrategy,
            capacity: int = 100,
            output_type: TypeInformation = None) -> 'DataStream':
        """
        Adds an async function with an AsyncRetryStrategy to support retry of AsyncFunction to the
        data stream. The order to process input records is guaranteed to be the same as input ones.

        :param data_stream: The input data stream.
        :param async_function: The async function.
        :param timeout: The timeout for the asynchronous operation to complete.
        :param async_retry_strategy: The strategy of reattempt async i/o operation that can be
                                     triggered
        :param capacity: The max number of async i/o operation that can be triggered.
        :param output_type: The output data type.
        :return: The transformed DataStream.
        """
        AsyncDataStream._validate(data_stream, async_function, timeout, async_retry_strategy)

        from pyflink.fn_execution import flink_fn_execution_pb2
        j_python_data_stream_function_operator, j_output_type_info = \
            _get_one_input_stream_operator(
                data_stream,
                AsyncFunctionDescriptor(
                    async_function, timeout, capacity, async_retry_strategy,
                    AsyncFunctionDescriptor.OutputMode.ORDERED),
                flink_fn_execution_pb2.UserDefinedDataStreamFunction.PROCESS,  # type: ignore
                output_type)
        return DataStream(data_stream._j_data_stream.transform(
            "async wait operator",
            j_output_type_info,
            j_python_data_stream_function_operator))

    @staticmethod
    def _validate(data_stream: DataStream, async_function: AsyncFunction,
                  timeout: Time, async_retry_strategy: AsyncRetryStrategy) -> None:
        if not inspect.iscoroutinefunction(async_function.async_invoke):
            raise Exception("Method 'async_invoke' of class '%s' should be declared as 'async def'."
                            % type(async_function))

        if async_retry_strategy is None:
            raise Exception("Async retry strategy should not be None.")
        if (async_retry_strategy != async_retry_strategies.NO_RETRY_STRATEGY and
                timeout.to_milliseconds() <= 0):
            raise Exception("Timeout should be configured when do async with retry.")
        if async_retry_strategy.get_retry_predicate() is None:
            raise Exception("Retry predicate of the async retry strategy '%s' is None."
                            % type(async_retry_strategy))

        gateway = get_gateway()
        j_conf = get_j_env_configuration(data_stream._j_data_stream.getExecutionEnvironment())
        python_execution_mode = (
            j_conf.get(gateway.jvm.org.apache.flink.python.PythonOptions.PYTHON_EXECUTION_MODE))
        if python_execution_mode == 'thread':
            raise Exception("AsyncFunction is still not supported for 'thread' mode.")



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from apache_beam.metrics.cells import DistributionData
from apache_beam.metrics.execution import MetricsContainer
from apache_beam.metrics.execution import MetricsEnvironment
from apache_beam.metrics.metricbase import MetricName
from apache_beam.runners.worker import statesampler
from apache_beam.utils import counters

from pyflink.fn_execution.metrics.process.metric_impl import GenericMetricGroup
from pyflink.metrics.metricbase import MetricGroup
from pyflink.table import FunctionContext
from pyflink.testing.test_case_utils import PyFlinkTestCase


class MetricTests(PyFlinkTestCase):

    base_metric_group = GenericMetricGroup(None, None)

    @staticmethod
    def print_metric_group_path(mg: MetricGroup) -> str:
        if mg._parent is None:
            return 'root'
        else:
            return MetricTests.print_metric_group_path(mg._parent) + '.' + mg._name

    def test_add_group(self):
        new_group = MetricTests.base_metric_group.add_group('my_group')
        self.assertEqual(MetricTests.print_metric_group_path(new_group), 'root.my_group')

    def test_add_group_with_variable(self):
        new_group = MetricTests.base_metric_group.add_group('key', 'value')
        self.assertEqual(MetricTests.print_metric_group_path(new_group), 'root.key.value')

    def test_metric_not_enabled(self):
        fc = FunctionContext(None, None)
        with self.assertRaises(RuntimeError):
            fc.get_metric_group()

    def test_function_context_runtime_info(self):
        fc = FunctionContext(
            None, {},
            task_name='MyTask',
            task_name_with_subtasks='MyTask (1/4)',
            number_of_parallel_subtasks=4,
            max_number_of_parallel_subtasks=128,
            index_of_this_subtask=0,
            attempt_number=2)
        self.assertEqual('MyTask', fc.get_task_name())
        self.assertEqual('MyTask (1/4)', fc.get_task_name_with_subtasks())
        self.assertEqual(4, fc.get_number_of_parallel_subtasks())
        self.assertEqual(128, fc.get_max_number_of_parallel_subtasks())
        self.assertEqual(0, fc.get_index_of_this_subtask())
        self.assertEqual(2, fc.get_attempt_number())

    def test_function_context_runtime_info_defaults(self):
        fc = FunctionContext(None, {})
        self.assertIsNone(fc.get_task_name())
        self.assertIsNone(fc.get_task_name_with_subtasks())
        self.assertIsNone(fc.get_number_of_parallel_subtasks())
        self.assertIsNone(fc.get_max_number_of_parallel_subtasks())
        self.assertIsNone(fc.get_index_of_this_subtask())
        self.assertIsNone(fc.get_attempt_number())

    def test_get_metric_name(self):
        new_group = MetricTests.base_metric_group.add_group('my_group')
        self.assertEqual(
            '["my_group", "MetricGroupType.generic"]',
            new_group._get_namespace())
        self.assertEqual(
            '["my_group", "MetricGroupType.generic", "60"]',
            new_group._get_namespace('60'))

    def test_metrics(self):
        sampler = statesampler.StateSampler('', counters.CounterFactory())
        statesampler.set_current_tracker(sampler)
        state1 = sampler.scoped_state(
            'mystep', 'myState', metrics_container=MetricsContainer('mystep'))

        try:
            sampler.start()
            with state1:
                counter = MetricTests.base_metric_group.counter("my_counter")
                meter = MetricTests.base_metric_group.meter("my_meter")
                distribution = MetricTests.base_metric_group.distribution("my_distribution")
                container = MetricsEnvironment.current_container()

                self.assertEqual(0, counter.get_count())
                self.assertEqual(0, meter.get_count())
                self.assertEqual(
                    DistributionData(
                        0, 0, 0, 0), container.get_distribution(
                        MetricName(
                            '[]', 'my_distribution')).get_cumulative())
                counter.inc(-2)
                meter.mark_event(3)
                distribution.update(10)
                distribution.update(2)
                self.assertEqual(-2, counter.get_count())
                self.assertEqual(3, meter.get_count())
                self.assertEqual(
                    DistributionData(
                        12, 2, 2, 10), container.get_distribution(
                        MetricName(
                            '[]', 'my_distribution')).get_cumulative())
        finally:
            sampler.stop()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.common import Duration
from pyflink.common.watermark_strategy import WatermarkStrategy
from pyflink.java_gateway import get_gateway
from pyflink.util.java_utils import is_instance_of, get_field_value

from pyflink.testing.test_case_utils import PyFlinkTestCase


class WatermarkStrategyTests(PyFlinkTestCase):

    def test_with_idleness(self):
        jvm = get_gateway().jvm
        j_watermark_strategy = WatermarkStrategy.no_watermarks().with_idleness(
            Duration.of_seconds(5)
        )._j_watermark_strategy
        self.assertTrue(is_instance_of(
            j_watermark_strategy,
            jvm.org.apache.flink.api.common.eventtime.WatermarkStrategyWithIdleness
        ))
        self.assertEqual(get_field_value(j_watermark_strategy, "idlenessTimeout").toMillis(), 5000)

    def test_with_watermark_alignment(self):
        jvm = get_gateway().jvm
        j_watermark_strategy = WatermarkStrategy.no_watermarks().with_watermark_alignment(
            "alignment-group-1", Duration.of_seconds(20), Duration.of_seconds(10)
        )._j_watermark_strategy
        self.assertTrue(is_instance_of(
            j_watermark_strategy,
            jvm.org.apache.flink.api.common.eventtime.WatermarksWithWatermarkAlignment
        ))
        alignment_parameters = j_watermark_strategy.getAlignmentParameters()
        self.assertEqual(alignment_parameters.getWatermarkGroup(), "alignment-group-1")
        self.assertEqual(alignment_parameters.getMaxAllowedWatermarkDrift(), 20000)
        self.assertEqual(alignment_parameters.getUpdateInterval(), 10000)

    def test_for_monotonous_timestamps(self):
        jvm = get_gateway().jvm
        j_watermark_strategy = WatermarkStrategy.for_monotonous_timestamps()._j_watermark_strategy
        self.assertTrue(is_instance_of(
            j_watermark_strategy.createWatermarkGenerator(None),
            jvm.org.apache.flink.api.common.eventtime.AscendingTimestampsWatermarks
        ))

    def test_for_bounded_out_of_orderness(self):
        jvm = get_gateway().jvm
        j_watermark_strategy = WatermarkStrategy.for_bounded_out_of_orderness(
            Duration.of_seconds(3)
        )._j_watermark_strategy
        j_watermark_generator = j_watermark_strategy.createWatermarkGenerator(None)
        self.assertTrue(is_instance_of(
            j_watermark_generator,
            jvm.org.apache.flink.api.common.eventtime.BoundedOutOfOrdernessWatermarks
        ))
        self.assertEqual(get_field_value(j_watermark_generator, "outOfOrdernessMillis"), 3000)

    def test_no_watermarks(self):
        jvm = get_gateway().jvm
        j_watermark_strategy = WatermarkStrategy.no_watermarks()._j_watermark_strategy
        self.assertTrue(is_instance_of(
            j_watermark_strategy.createWatermarkGenerator(None),
            jvm.org.apache.flink.api.common.eventtime.NoWatermarksGenerator
        ))



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.table import DataTypes
from pyflink.table.expressions import call_sql
from pyflink.table.schema import Schema
from pyflink.testing.test_case_utils import PyFlinkTestCase


class SchemaTest(PyFlinkTestCase):
    def test_schema_basic(self):
        old_schema = Schema.new_builder() \
            .from_row_data_type(DataTypes.ROW(
                [DataTypes.FIELD("a", DataTypes.TINYINT()),
                 DataTypes.FIELD("b", DataTypes.SMALLINT()),
                 DataTypes.FIELD("c", DataTypes.INT())])) \
            .from_fields(["d", "e"], [DataTypes.STRING(), DataTypes.BOOLEAN()]) \
            .build()
        self.schema = Schema.new_builder() \
            .from_schema(old_schema) \
            .primary_key_named("primary_constraint", "id") \
            .column("id", DataTypes.INT().not_null()) \
            .column("counter", DataTypes.INT().not_null()) \
            .column("payload", "ROW<name STRING, age INT, flag BOOLEAN>") \
            .column_by_metadata("topic", DataTypes.STRING(), None, True) \
            .column_by_expression("ts", call_sql("orig_ts - INTERVAL '60' MINUTE")) \
            .column_by_metadata("orig_ts", DataTypes.TIMESTAMP(3), "timestamp") \
            .watermark("ts", "ts - INTERVAL '5' SECOND") \
            .column_by_expression("proctime", "PROCTIME()") \
            .build()
        self.assertEqual("""(
  `a` TINYINT,
  `b` SMALLINT,
  `c` INT,
  `d` STRING,
  `e` BOOLEAN,
  `id` INT NOT NULL,
  `counter` INT NOT NULL,
  `payload` [ROW<name STRING, age INT, flag BOOLEAN>],
  `topic` STRING METADATA VIRTUAL,
  `ts` AS [orig_ts - INTERVAL '60' MINUTE],
  `orig_ts` TIMESTAMP(3) METADATA FROM 'timestamp',
  `proctime` AS [PROCTIME()],
  WATERMARK FOR `ts` AS [ts - INTERVAL '5' SECOND],
  CONSTRAINT `primary_constraint` PRIMARY KEY (`id`) NOT ENFORCED
)""", str(self.schema))


if __name__ == '__main__':
    import unittest

    try:
        import xmlrunner

        testRunner = xmlrunner.XMLTestRunner(output='target/test-reports')
    except ImportError:
        testRunner = None
    unittest.main(testRunner=testRunner, verbosity=2)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
# #  distributed under the License is distributed on an "AS IS" BASIS,
# #  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# #  See the License for the specific language governing permissions and
# # limitations under the License.
################################################################################
import datetime
import decimal
import sys
import unittest

from py4j.protocol import Py4JJavaError
from typing import Iterable

from pyflink.common import RowKind, WatermarkStrategy, Configuration
from pyflink.common.serializer import TypeSerializer
from pyflink.common.typeinfo import Types
from pyflink.common.watermark_strategy import TimestampAssigner
from pyflink.datastream import MergingWindowAssigner, TimeWindow, Trigger, TriggerResult, OutputTag
from pyflink.datastream.functions import WindowFunction, ProcessFunction
from pyflink.datastream.tests.test_util import DataStreamTestSinkFunction
from pyflink.datastream.window import TimeWindowSerializer
from pyflink.java_gateway import get_gateway
from pyflink.table import (DataTypes, StreamTableEnvironment, EnvironmentSettings, Module,
                           ResultKind, ModuleEntry, Schema)
from pyflink.table.catalog import ObjectPath, CatalogBaseTable, CatalogDescriptor
from pyflink.table.explain_detail import ExplainDetail
from pyflink.table.expressions import col, source_watermark
from pyflink.table.table_descriptor import TableDescriptor
from pyflink.table.types import RowType, Row, UserDefinedType
from pyflink.table.udf import udf
from pyflink.testing import source_sink_utils
from pyflink.testing.test_case_utils import (PyFlinkStreamTableTestCase, PyFlinkUTTestCase,
                                             _load_specific_flink_module_jars)
from pyflink.util.java_utils import get_j_env_configuration


class TableEnvironmentTest(PyFlinkUTTestCase):

    def test_set_sys_executable_for_local_mode(self):
        jvm = get_gateway().jvm
        actual_executable = get_j_env_configuration(self.t_env._get_j_env()) \
            .getString(jvm.PythonOptions.PYTHON_EXECUTABLE.key(), None)
        self.assertEqual(sys.executable, actual_executable)

    def test_explain(self):
        schema = RowType() \
            .add('a', DataTypes.INT()) \
            .add('b', DataTypes.STRING()) \
            .add('c', DataTypes.STRING())
        t_env = self.t_env
        t = t_env.from_elements([], schema)
        result = t.select(t.a + 1, t.b, t.c)

        actual = result.explain()

        assert isinstance(actual, str)

    def test_explain_sql(self):
        t_env = self.t_env
        actual = t_env.explain_sql("SELECT * FROM (VALUES ('a', 1))")
        assert isinstance(actual, str)

    def test_explain_with_extended(self):
        schema = RowType() \
            .add('a', DataTypes.INT()) \
            .add('b', DataTypes.STRING()) \
            .add('c', DataTypes.STRING())
        t_env = self.t_env
        t = t_env.from_elements([], schema)
        result = t.select(t.a + 1, t.b, t.c)

        actual = result.explain(ExplainDetail.ESTIMATED_COST, ExplainDetail.CHANGELOG_MODE,
                                ExplainDetail.JSON_EXECUTION_PLAN, ExplainDetail.PLAN_ADVICE)

        assert isinstance(actual, str)

    def test_explain_sql_extended(self):
        t_env = self.t_env
        actual = t_env.explain_sql(
            "SELECT * FROM (VALUES ('a', 1))",
            ExplainDetail.ESTIMATED_COST,
            ExplainDetail.CHANGELOG_MODE,
            ExplainDetail.JSON_EXECUTION_PLAN,
            ExplainDetail.PLAN_ADVICE
        )
        assert isinstance(actual, str)

    def test_register_functions(self):
        t_env = self.t_env

        t_env.create_temporary_system_function(
            "python_scalar_func", udf(lambda i: i, result_type=DataTypes.INT()))

        t_env.create_java_temporary_system_function(
            "scalar_func", "org.apache.flink.table.utils.TestingFunctions$RichFunc0")

        t_env.create_java_temporary_system_function(
            "agg_func", "org.apache.flink.table.utils.TestingFunctions$ByteMaxAggFunction")
        t_env.create_java_temporary_system_function(
            "table_func", "org.apache.flink.table.utils.TestingFunctions$TableFunc1")

        actual = t_env.list_user_defined_functions()
        expected = ['python_scalar_func', 'scalar_func', 'agg_func', 'table_func']
        self.assert_equals(actual, expected)

    def test_load_module_twice(self):
        t_env = self.t_env
        self.check_list_modules('core')
        self.check_list_full_modules(1, 'core')
        self.assertRaisesRegex(
            Py4JJavaError, "A module with name 'core' already exists",
            t_env.load_module, 'core', Module(
                get_gateway().jvm.org.apache.flink.table.module.CoreModule.INSTANCE))

    def test_unload_module_twice(self):
        t_env = self.t_env
        t_env.unload_module('core')
        self.check_list_modules()
        self.check_list_full_modules(0)
        self.assertRaisesRegex(
            Py4JJavaError, "No module with name 'core' exists",
            t_env.unload_module, 'core')

    def test_use_modules(self):
        # please do not change this order since ModuleMock depends on FunctionDefinitionMock
        _load_specific_flink_module_jars('/flink-table/flink-table-common')
        _load_specific_flink_module_jars('/flink-table/flink-table-api-java')

        t_env = self.t_env
        t_env.load_module('x', Module(
            get_gateway().jvm.org.apache.flink.table.utils.ModuleMock("x")
        ))
        t_env.load_module('y', Module(
            get_gateway().jvm.org.apache.flink.table.utils.ModuleMock("y")
        ))
        self.check_list_modules('core', 'x', 'y')
        self.check_list_full_modules(3, 'core', 'x', 'y')

        t_env.use_modules('y', 'core')
        self.check_list_modules('y', 'core')
        self.check_list_full_modules(2, 'y', 'core', 'x')

    def check_list_modules(self, *expected_used_modules: str):
        self.assert_equals(self.t_env.list_modules(), list(expected_used_modules))

    def check_list_full_modules(self, used_module_cnt: int, *expected_loaded_modules: str):
        self.assert_equals(self.t_env.list_full_modules(),
                           [ModuleEntry(module,
                                        expected_loaded_modules.index(module) < used_module_cnt)
                            for module in expected_loaded_modules])

    def test_unload_and_load_module(self):
        t_env = self.t_env
        t_env.unload_module('core')
        t_env.load_module('core', Module(
            get_gateway().jvm.org.apache.flink.table.module.CoreModule.INSTANCE))
        table_result = t_env.execute_sql("select concat('unload', 'load') as test_module")
        self.assertEqual(table_result.get_result_kind(), ResultKind.SUCCESS_WITH_CONTENT)
        self.assert_equals(table_result.get_resolved_schema().get_column_names(), ['test_module'])

    def test_create_and_drop_java_function(self):
        t_env = self.t_env

        t_env.create_java_temporary_system_function(
            "scalar_func", "org.apache.flink.table.utils.TestingFunctions$RichFunc0")
        t_env.create_java_function(
            "agg_func", "org.apache.flink.table.utils.TestingFunctions$ByteMaxAggFunction")
        t_env.create_java_temporary_function(
            "table_func", "org.apache.flink.table.utils.TestingFunctions$TableFunc1")
        self.assert_equals(t_env.list_user_defined_functions(),
                           ['scalar_func', 'agg_func', 'table_func'])

        t_env.drop_temporary_system_function("scalar_func")
        t_env.drop_function("agg_func")
        t_env.drop_temporary_function("table_func")
        self.assert_equals(t_env.list_user_defined_functions(), [])

    def test_create_temporary_table_from_descriptor(self):
        schema = Schema.new_builder().column("f0", DataTypes.INT()).build()

        self.assert_created_temporary_table_from_descriptor(schema)

    def test_create_temporary_table_if_not_exists_from_descriptor(self):
        schema = Schema.new_builder().column("f0", DataTypes.INT()).build()

        self.assert_created_temporary_table_from_descriptor(schema)

        # This should be a no-op and throw no exception
        self.t_env.create_temporary_table(
            "T",
            TableDescriptor.for_connector("fake")
            .schema(schema)
            .option("a", "Test")
            .build(),
            True)

        with self.assertRaises(Exception) as error_context:
            self.t_env.create_temporary_table(
                "T",
                TableDescriptor.for_connector("fake")
                .schema(schema)
                .option("a", "Test")
                .build(),
                False)

        error_msg = str(error_context.exception)
        self.assertIn("Temporary table '`default_catalog`.`default_database`.`T`' already exists",
                      error_msg)

    def test_create_table_from_descriptor(self):
        schema = Schema.new_builder().column("f0", DataTypes.INT()).build()

        self.assert_created_table_from_descriptor(schema)

    def test_create_table_ignore_if_exists_from_descriptor(self):
        schema = Schema.new_builder().column("f0", DataTypes.INT()).build()
        self.assert_created_table_from_descriptor(schema)

        # This should be a no-op and throw no exception
        self.t_env.create_table(
            "T",
            TableDescriptor.for_connector("fake")
            .schema(schema)
            .option("a", "Test")
            .build(),
            True)

        with self.assertRaises(Exception) as error_context:
            self.t_env.create_table(
                "T",
                TableDescriptor.for_connector("fake")
                .schema(schema)
                .option("a", "Test")
                .build(),
                False)

        error_msg = str(error_context.exception)
        self.assertIn(
            "Could not execute CreateTable in path `default_catalog`.`default_database`.`T`",
            error_msg)

    def test_table_from_descriptor(self):
        schema = Schema.new_builder().column("f0", DataTypes.INT()).build()
        descriptor = TableDescriptor.for_connector("fake").schema(schema).build()

        table = self.t_env.from_descriptor(descriptor)
        self.assertEqual(schema,
                         Schema.new_builder().from_resolved_schema(
                             table.get_resolved_schema()
                         ).build())
        contextResolvedTable = table._j_table.getQueryOperation().getContextResolvedTable()
        options = contextResolvedTable.getTable().getOptions()
        self.assertEqual("fake", options.get("connector"))

    def test_udt(self):
        self.t_env.from_elements([
            (DenseVector([1, 2, 3, 4]), 0., 1.),
            (DenseVector([2, 2, 3, 4]), 0., 2.),
            (DenseVector([3, 2, 3, 4]), 0., 3.),
            (DenseVector([4, 2, 3, 4]), 0., 4.),
            (DenseVector([5, 2, 3, 4]), 0., 5.),
            (DenseVector([11, 2, 3, 4]), 1., 1.),
            (DenseVector([12, 2, 3, 4]), 1., 2.),
            (DenseVector([13, 2, 3, 4]), 1., 3.),
            (DenseVector([14, 2, 3, 4]), 1., 4.),
            (DenseVector([15, 2, 3, 4]), 1., 5.),
        ],
            DataTypes.ROW([
                DataTypes.FIELD("features", VectorUDT()),
                DataTypes.FIELD("label", DataTypes.DOUBLE()),
                DataTypes.FIELD("weight", DataTypes.DOUBLE())]))

    def test_explain_with_multi_sinks(self):
        t_env = self.t_env
        source = t_env.from_elements([(1, "Hi", "Hello"), (2, "Hello", "Hello")], ["a", "b", "c"])
        t_env.execute_sql("""
            CREATE TABLE sink1 (
                a BIGINT,
                b STRING,
                c STRING
            ) WITH (
                'connector' = 'filesystem',
                'path'='path1',
                'format' = 'csv'
            )
        """)

        t_env.execute_sql("""
            CREATE TABLE sink2 (
                a BIGINT,
                b STRING,
                c STRING
            ) WITH (
                'connector' = 'filesystem',
                'path'='path2',
                'format' = 'csv'
            )
        """)

        stmt_set = t_env.create_statement_set()
        stmt_set.add_insert_sql("insert into sink1 select * from %s where a > 100" % source)
        stmt_set.add_insert_sql("insert into sink2 select * from %s where a < 100" % source)

        actual = stmt_set.explain(ExplainDetail.ESTIMATED_COST, ExplainDetail.CHANGELOG_MODE,
                                  ExplainDetail.JSON_EXECUTION_PLAN)
        self.assertIsInstance(actual, str)

    def test_register_java_function(self):
        t_env = self.t_env

        t_env.create_java_temporary_system_function(
            "scalar_func", "org.apache.flink.table.utils.TestingFunctions$RichFunc0")

        t_env.create_java_temporary_system_function(
            "agg_func", "org.apache.flink.table.utils.TestingFunctions$ByteMaxAggFunction")

        t_env.create_java_temporary_system_function(
            "table_func", "org.apache.flink.table.utils.TestingFunctions$TableFunc1")

        actual = t_env.list_user_defined_functions()
        expected = ['scalar_func', 'agg_func', 'table_func']
        self.assert_equals(actual, expected)

    def test_use_duplicated_modules(self):
        self.assertRaisesRegex(
            Py4JJavaError, "Module 'core' appears more than once",
            self.t_env.use_modules, 'core', 'core')

    def test_use_nonexistent_module(self):
        self.assertRaisesRegex(
            Py4JJavaError, "No module with name 'dummy' exists",
            self.t_env.use_modules, 'core', 'dummy')

    def assert_created_table_from_descriptor(self, schema: Schema):
        catalog = self.t_env.get_current_catalog()
        database = self.t_env.get_current_database()

        self.t_env.create_table(
            "T",
            TableDescriptor.for_connector("fake")
            .schema(schema)
            .option("a", "Test")
            .build())

        object_path = ObjectPath(database, "T")
        self.assertTrue(self.t_env.get_catalog(catalog).table_exists(object_path))

        catalog_table = self.t_env.get_catalog(catalog).get_table(object_path)
        self.assertEqual(schema, catalog_table.get_unresolved_schema())
        self.assertEqual("fake", catalog_table.get_options().get("connector"))
        self.assertEqual("Test", catalog_table.get_options().get("a"))

    def assert_created_temporary_table_from_descriptor(self, schema: Schema):
        catalog = self.t_env.get_current_catalog()
        database = self.t_env.get_current_database()

        self.t_env.create_temporary_table(
            "T",
            TableDescriptor.for_connector("fake")
            .schema(schema)
            .option("a", "Test")
            .build())

        self.assertFalse(self.t_env.get_catalog(catalog).table_exists(ObjectPath(database, "T")))
        gateway = get_gateway()

        catalog_table = CatalogBaseTable(
            self.t_env._j_tenv.getCatalogManager()
            .getTable(gateway.jvm.ObjectIdentifier.of(catalog, database, "T"))
            .get()
            .getTable())
        self.assertEqual(schema, catalog_table.get_unresolved_schema())
        self.assertEqual("fake", catalog_table.get_options().get("connector"))
        self.assertEqual("Test", catalog_table.get_options().get("a"))


class DataStreamConversionTestCases(PyFlinkUTTestCase):

    def setUp(self) -> None:
        from pyflink.datastream import StreamExecutionEnvironment

        super(DataStreamConversionTestCases, self).setUp()
        config = Configuration()
        config.set_string("pekko.ask.timeout", "20 s")
        self.env = StreamExecutionEnvironment.get_execution_environment(config)
        self.t_env = StreamTableEnvironment.create(self.env)

        self.env.set_parallelism(2)
        self.t_env.get_config().set(
            "python.fn-execution.bundle.size", "1")
        self.test_sink = DataStreamTestSinkFunction()

    def test_from_data_stream_atomic(self):
        data_stream = self.env.from_collection([(1,), (2,), (3,), (4,), (5,)])
        result = self.t_env.from_data_stream(data_stream).execute()
        self.assertEqual("""(
  `f0` RAW('[B', '...')
)""",
                         str(result.get_resolved_schema()))
        with result.collect() as result:
            collected_result = [str(item) for item in result]
            expected_result = [item for item
                               in map(str, [Row((1,)), Row((2,)), Row((3,)), Row((4,)), Row((5,))])]
            expected_result.sort()
            collected_result.sort()
            self.assertEqual(expected_result, collected_result)

    def test_to_data_stream_atomic(self):
        table = self.t_env.from_elements([(1,), (2,), (3,)], ["a"])
        ds = self.t_env.to_data_stream(table)
        ds.add_sink(self.test_sink)
        self.env.execute()
        results = self.test_sink.get_results(False)
        results.sort()
        expected = ['+I[1]', '+I[2]', '+I[3]']
        self.assertEqual(expected, results)

    def test_to_data_stream_local_time(self):
        self.t_env.execute_sql("""
        CREATE TEMPORARY VIEW v0 AS
        SELECT f0, f1, f2, f3 FROM ( VALUES
            ( 1, DATE'1970-01-02', TIME'03:04:05', TIMESTAMP'1970-01-02 03:04:05' ),
            ( 2, DATE'1970-06-07', TIME'08:09:10', TIMESTAMP'1970-06-07 08:09:10' )
        ) AS t0 ( f0, f1, f2, f3 )
        """)
        v0 = self.t_env.from_path("v0")
        self.t_env.to_data_stream(v0).key_by(lambda r: r['f0']).add_sink(self.test_sink)
        self.env.execute()
        results = self.test_sink.get_results(False)
        results.sort()
        expected = ['+I[1, 1970-01-02, 03:04:05, 1970-01-02T03:04:05]',
                    '+I[2, 1970-06-07, 08:09:10, 1970-06-07T08:09:10]']
        self.assertEqual(expected, results)

    def test_from_data_stream(self):
        self.env.set_parallelism(1)

        ds = self.env.from_collection([(1, 'Hi', 'Hello'), (2, 'Hello', 'Hi')],
                                      type_info=Types.ROW([Types.INT(),
                                                           Types.STRING(),
                                                           Types.STRING()]))
        t_env = self.t_env
        table = t_env.from_data_stream(ds)
        sink_table_ddl = """
        CREATE TABLE Sink(a INT, b STRING, c STRING) WITH ('connector'='test-sink')
        """
        t_env.execute_sql(sink_table_ddl)
        expr_sink_ddl = """
        CREATE TABLE ExprSink(a INT, b STRING, c STRING) WITH ('connector'='test-sink')
        """
        t_env.execute_sql(expr_sink_ddl)
        table.execute_insert("Sink").wait()
        result = source_sink_utils.results()
        expected = ['+I[1, Hi, Hello]', '+I[2, Hello, Hi]']
        self.assert_equals(result, expected)

        ds = ds.map(lambda x: x, Types.ROW([Types.INT(), Types.STRING(), Types.STRING()])) \
               .map(lambda x: x, Types.ROW([Types.INT(), Types.STRING(), Types.STRING()]))
        table = t_env.from_data_stream(ds, col('a'), col('b'), col('c'))
        table.execute_insert("ExprSink").wait()
        result = source_sink_utils.results()
        self.assert_equals(result, expected)

    def test_from_data_stream_with_schema(self):
        ds = self.env.from_collection([(1, 'Hi', 'Hello'), (2, 'Hello', 'Hi')],
                                      type_info=Types.ROW_NAMED(
                                          ["a", "b", "c"],
                                          [Types.INT(), Types.STRING(), Types.STRING()]))

        table = self.t_env.from_data_stream(ds,
                                            Schema.new_builder()
                                                  .column("a", DataTypes.INT())
                                                  .column("b", DataTypes.STRING())
                                                  .column("c", DataTypes.STRING())
                                                  .build())
        result = table.execute()
        with result.collect() as result:
            collected_result = [str(item) for item in result]
            expected_result = [item for item in
                               map(str, [Row(1, 'Hi', 'Hello'), Row(2, 'Hello', 'Hi')])]
            expected_result.sort()
            collected_result.sort()
            self.assertEqual(expected_result, collected_result)

    @unittest.skip
    def test_from_and_to_data_stream_event_time(self):
        from pyflink.table import Schema

        ds = self.env.from_collection([(1, 42, "a"), (2, 5, "a"), (3, 1000, "c"), (100, 1000, "c")],
                                      Types.ROW_NAMED(
                                          ["a", "b", "c"],
                                          [Types.LONG(), Types.INT(), Types.STRING()]))
        ds = ds.assign_timestamps_and_watermarks(
            WatermarkStrategy.for_monotonous_timestamps()
            .with_timestamp_assigner(MyTimestampAssigner()))

        table = self.t_env.from_data_stream(ds,
                                            Schema.new_builder()
                                                  .column_by_metadata("rowtime", "TIMESTAMP_LTZ(3)")
                                                  .watermark("rowtime", "SOURCE_WATERMARK()")
                                                  .build())
        self.assertEqual("""(
  `a` BIGINT,
  `b` INT,
  `c` STRING,
  `rowtime` TIMESTAMP_LTZ(3) *ROWTIME* METADATA,
  WATERMARK FOR `rowtime`: TIMESTAMP_LTZ(3) AS SOURCE_WATERMARK()
)""",
                         str(table.get_resolved_schema()))
        self.t_env.create_temporary_view("t",
                                         ds,
                                         Schema.new_builder()
                                         .column_by_metadata("rowtime", "TIMESTAMP_LTZ(3)")
                                         .watermark("rowtime", "SOURCE_WATERMARK()")
                                         .build())

        result = self.t_env.execute_sql("SELECT "
                                        "c, SUM(b) "
                                        "FROM t "
                                        "GROUP BY c, TUMBLE(rowtime, INTERVAL '0.005' SECOND)")
        with result.collect() as result:
            collected_result = [str(item) for item in result]
            expected_result = [item for item in
                               map(str, [Row('a', 47), Row('c', 1000), Row('c', 1000)])]
            expected_result.sort()
            collected_result.sort()
            self.assertEqual(expected_result, collected_result)

        ds = self.t_env.to_data_stream(table)
        ds.key_by(lambda k: k.c, key_type=Types.STRING()) \
            .window(MyTumblingEventTimeWindow()) \
            .apply(SumWindowFunction(), Types.TUPLE([Types.STRING(), Types.INT()])) \
            .add_sink(self.test_sink)
        self.env.execute()
        expected_results = ['(a,47)', '(c,1000)', '(c,1000)']
        actual_results = self.test_sink.get_results(False)
        expected_results.sort()
        actual_results.sort()
        self.assertEqual(expected_results, actual_results)

    def test_from_and_to_changelog_stream_event_time(self):
        from pyflink.table import Schema

        self.env.set_parallelism(1)
        ds = self.env.from_collection([(1, 42, "a"), (2, 5, "a"), (3, 1000, "c"), (100, 1000, "c")],
                                      Types.ROW([Types.LONG(), Types.INT(), Types.STRING()]))
        ds = ds.assign_timestamps_and_watermarks(
            WatermarkStrategy.for_monotonous_timestamps()
            .with_timestamp_assigner(MyTimestampAssigner()))

        changelog_stream = ds.map(lambda t: Row(t.f1, t.f2),
                                  Types.ROW([Types.INT(), Types.STRING()]))

        # derive physical columns and add a rowtime
        table = self.t_env.from_changelog_stream(
            changelog_stream,
            Schema.new_builder()
                  .column_by_metadata("rowtime", DataTypes.TIMESTAMP_LTZ(3))
                  .column_by_expression("computed", str(col("f1").upper_case))
                  .watermark("rowtime", str(source_watermark()))
                  .build())

        self.t_env.create_temporary_view("t", table)

        # access and reorder columns
        reordered = self.t_env.sql_query("SELECT computed, rowtime, f0 FROM t")

        # write out the rowtime column with fully declared schema
        result = self.t_env.to_changelog_stream(
            reordered,
            Schema.new_builder()
            .column("f1", DataTypes.STRING())
            .column_by_metadata("rowtime", DataTypes.TIMESTAMP_LTZ(3))
            .column_by_expression("ignored", str(col("f1").upper_case))
            .column("f0", DataTypes.INT())
            .build()
        )

        # test event time window and field access
        result.key_by(lambda k: k.f1) \
            .window(MyTumblingEventTimeWindow()) \
            .apply(SumWindowFunction(), Types.TUPLE([Types.STRING(), Types.INT()])) \
            .add_sink(self.test_sink)
        self.env.execute()
        expected_results = ['(A,47)', '(C,1000)', '(C,1000)']
        actual_results = self.test_sink.get_results(False)
        expected_results.sort()
        actual_results.sort()
        self.assertEqual(expected_results, actual_results)

    def test_to_append_stream(self):
        self.env.set_parallelism(1)
        t_env = StreamTableEnvironment.create(
            self.env,
            environment_settings=EnvironmentSettings.in_streaming_mode())
        table = t_env.from_elements([(1, "Hi", "Hello"), (2, "Hello", "Hi")], ["a", "b", "c"])
        new_table = table.select(table.a + 1, table.b + 'flink', table.c)
        ds = t_env.to_append_stream(table=new_table, type_info=Types.ROW([Types.LONG(),
                                                                          Types.STRING(),
                                                                          Types.STRING()]))
        test_sink = DataStreamTestSinkFunction()
        ds.add_sink(test_sink)
        self.env.execute("test_to_append_stream")
        result = test_sink.get_results(False)
        expected = ['+I[2, Hiflink, Hello]', '+I[3, Helloflink, Hi]']
        self.assertEqual(result, expected)

    def test_to_retract_stream(self):
        self.env.set_parallelism(1)
        t_env = StreamTableEnvironment.create(
            self.env,
            environment_settings=EnvironmentSettings.in_streaming_mode())
        table = t_env.from_elements([(1, "Hi", "Hello"), (1, "Hi", "Hello")], ["a", "b", "c"])
        new_table = table.group_by(table.c).select(table.a.sum, table.c.alias("b"))
        ds = t_env.to_retract_stream(table=new_table, type_info=Types.ROW([Types.LONG(),
                                                                           Types.STRING()]))
        test_sink = DataStreamTestSinkFunction()
        ds.map(lambda x: x).add_sink(test_sink)
        self.env.execute("test_to_retract_stream")
        result = test_sink.get_results(True)
        expected = ["(True, Row(f0=1, f1='Hello'))", "(False, Row(f0=1, f1='Hello'))",
                    "(True, Row(f0=2, f1='Hello'))"]
        self.assertEqual(result, expected)

    def test_side_output_stream_to_table(self):
        tag = OutputTag("side", Types.ROW([Types.INT()]))

        class MyProcessFunction(ProcessFunction):

            def process_element(self, value, ctx):
                yield Row(value)
                yield tag, Row(value * 2)

        ds = self.env.from_collection([1, 2, 3], Types.INT()).process(MyProcessFunction())
        ds_side = ds.get_side_output(tag)
        expected = ['<Row(2)>', '<Row(4)>', '<Row(6)>']

        t = self.t_env.from_data_stream(ds_side)
        result = [str(i) for i in t.execute().collect()]
        result.sort()
        self.assertEqual(expected, result)

        self.t_env.create_temporary_view("side_table", ds_side)
        table_result = self.t_env.execute_sql("SELECT * FROM side_table")
        result = [str(i) for i in table_result.collect()]
        result.sort()
        self.assertEqual(expected, result)

    def test_create_catalog(self):
        config = Configuration()
        config.set_string("type", "generic_in_memory")
        catalog_desc = CatalogDescriptor.of("mycat", config, None)
        self.t_env.create_catalog("mycat", catalog_desc)
        mycat = self.t_env.get_catalog("mycat")
        self.assertIsNotNone(mycat)


class StreamTableEnvironmentTests(PyFlinkStreamTableTestCase):

    def test_collect_with_retract(self):
        expected_row_kinds = [RowKind.INSERT, RowKind.UPDATE_BEFORE, RowKind.UPDATE_AFTER,
                              RowKind.INSERT, RowKind.UPDATE_BEFORE, RowKind.UPDATE_AFTER]
        element_data = [(1, 2, 'a'),
                        (3, 4, 'b'),
                        (5, 6, 'a'),
                        (7, 8, 'b')]
        field_names = ['a', 'b', 'c']
        source = self.t_env.from_elements(element_data, field_names)
        table_result = self.t_env.execute_sql(
            "SELECT SUM(a), c FROM %s group by c" % source)
        with table_result.collect() as result:
            collected_result = []
            for i in result:
                collected_result.append(i)

            collected_result = [str(result) + ',' + str(result.get_row_kind())
                                for result in collected_result]
            expected_result = [Row(1, 'a'), Row(1, 'a'), Row(6, 'a'), Row(3, 'b'),
                               Row(3, 'b'), Row(10, 'b')]
            for i in range(len(expected_result)):
                expected_result[i] = str(expected_result[i]) + ',' + str(expected_row_kinds[i])
            expected_result.sort()
            collected_result.sort()
            self.assertEqual(expected_result, collected_result)

    def test_collect_for_all_data_types(self):
        expected_result = [Row(1, None, 1, True, 32767, -2147483648, 1.23,
                               1.98932, bytearray(b'pyflink'), 'pyflink',
                               datetime.date(2014, 9, 13), datetime.time(12, 0, 0, 123000),
                               datetime.datetime(2018, 3, 11, 3, 0, 0, 123000),
                               [['a', 'b'], ['c', 'd'], ['e', 'f']],
                               [Row('pyflink'), Row('pyflink'), Row('pyflink')],
                               {1: Row('flink'), 2: Row('pyflink')},
                               [Row('a1', {1: Row('b1')}, [Row('c1', 'd1'), Row('e1', 'f1')]),
                                Row('a2', {2: Row('b2')}, [Row('c2', 'd2'), Row('e2', 'f2')])],
                               decimal.Decimal('1000000000000000000.050000000000000000'),
                               decimal.Decimal('1000000000000000000.059999999999999999'))]
        source = self.t_env.from_elements(
            [(1, None, 1, True, 32767, -2147483648, 1.23, 1.98932, bytearray(b'pyflink'), 'pyflink',
              datetime.date(2014, 9, 13), datetime.time(hour=12, minute=0, second=0,
                                                        microsecond=123000),
              datetime.datetime(2018, 3, 11, 3, 0, 0, 123000),
              [['a', 'b'], ['c', 'd'], ['e', 'f']],
              [Row('pyflink'), Row('pyflink'), Row('pyflink')],
              {1: Row('flink'), 2: Row('pyflink')},
              [Row('a1', {1: Row('b1')}, [Row('c1', 'd1'), Row('e1', 'f1')]),
               Row('a2', {2: Row('b2')}, [Row('c2', 'd2'), Row('e2', 'f2')])],
              decimal.Decimal('1000000000000000000.05'),
              decimal.Decimal('1000000000000000000.05999999999999999899999999999'))],
            DataTypes.ROW(
                [DataTypes.FIELD("a", DataTypes.BIGINT()),
                 DataTypes.FIELD("b", DataTypes.BIGINT()),
                 DataTypes.FIELD("c", DataTypes.TINYINT()),
                 DataTypes.FIELD("d", DataTypes.BOOLEAN()),
                 DataTypes.FIELD("e", DataTypes.SMALLINT()),
                 DataTypes.FIELD("f", DataTypes.INT()),
                 DataTypes.FIELD("g", DataTypes.FLOAT()),
                 DataTypes.FIELD("h", DataTypes.DOUBLE()),
                 DataTypes.FIELD("i", DataTypes.BYTES()),
                 DataTypes.FIELD("j", DataTypes.STRING()),
                 DataTypes.FIELD("k", DataTypes.DATE()),
                 DataTypes.FIELD("l", DataTypes.TIME()),
                 DataTypes.FIELD("m", DataTypes.TIMESTAMP(3)),
                 DataTypes.FIELD("n", DataTypes.ARRAY(DataTypes.ARRAY(DataTypes.STRING()))),
                 DataTypes.FIELD(
                     "o",
                     DataTypes.ARRAY(
                         DataTypes.ROW([DataTypes.FIELD("ss2", DataTypes.STRING())])
                     )),
                 DataTypes.FIELD(
                     "p",
                     DataTypes.MAP(
                         DataTypes.BIGINT(),
                         DataTypes.ROW([DataTypes.FIELD("ss", DataTypes.STRING())]),
                     )),
                 DataTypes.FIELD(
                     "q",
                     DataTypes.ARRAY(
                         DataTypes.ROW(
                             [
                                 DataTypes.FIELD("a1", DataTypes.STRING()),
                                 DataTypes.FIELD(
                                     "a2",
                                     DataTypes.MAP(
                                         DataTypes.BIGINT(),
                                         DataTypes.ROW(
                                             [DataTypes.FIELD("ss", DataTypes.STRING())]
                                         ),
                                     ),
                                 ),
                                 DataTypes.FIELD(
                                     "a3",
                                     DataTypes.ARRAY(
                                         DataTypes.ROW(
                                             [
                                                 DataTypes.FIELD("a1", DataTypes.STRING()),
                                                 DataTypes.FIELD("a2", DataTypes.STRING()),
                                             ]
                                         )
                                     ),
                                 ),
                             ]
                         )
                     )),
                 DataTypes.FIELD("r", DataTypes.DECIMAL(38, 18)),
                 DataTypes.FIELD("s", DataTypes.DECIMAL(38, 18))
                 ]
            )
        )
        table_result = source.execute()
        with table_result.collect() as result:
            collected_result = []
            for i in result:
                collected_result.append(i)
            self.assertEqual(expected_result, collected_result)

    def test_row_form_consistency_with_elements(self):
        schema = DataTypes.ROW(
            [
                DataTypes.FIELD(
                    "col",
                    DataTypes.ARRAY(
                        DataTypes.ROW(
                            [
                                DataTypes.FIELD("a", DataTypes.STRING()),
                                DataTypes.FIELD("b", DataTypes.BOOLEAN()),
                            ]
                        )
                    ),
                ),
            ]
        )

        valid_tuple_elements = [(
            [("pyflink", True), ("pyflink", False), ("pyflink", True)],
        )]
        valid_list_elements = [(
            [["pyflink", True], ["pyflink", False], ["pyflink", True]],
        )]
        valid_keyword_row = [(
            [Row(a="pyflink", b=True), Row(a="pyflink", b=False), Row(a="pyflink", b=True)],
        )]
        valid_positional_row = [(
            [Row("pyflink", True), Row("pyflink", False), Row("pyflink", True)],
        )]
        expected_valid_result = [
            Row([
                Row("pyflink", True), Row("pyflink", False), Row("pyflink", True)
            ])
        ]

        for elements in (
            valid_tuple_elements,
            valid_list_elements,
            valid_keyword_row,
            valid_positional_row
        ):
            table = self.t_env.from_elements(elements, schema)
            table_result = list(table.execute().collect())
            self.assertEqual(table_result, expected_valid_result)

        invalid_tuple_elements = [(
            [("pyflink", True), ("pyflink", False), (True, "pyflink")],
        )]
        invalid_list_elements = [(
            [["pyflink", True], ["pyflink", False], [True, "pyflink"]],
        )]
        invalid_keyword_row = [(
            [Row(a="pyflink", b=True), Row(a="pyflink", b=False), Row(a=True, b="pyflink")],
        )]
        invalid_positional_row = [(
            [Row("pyflink", True), Row("pyflink", False), Row(True, "pyflink")],
        )]

        for elements in (
            invalid_tuple_elements,
            invalid_list_elements,
            invalid_keyword_row,
            invalid_positional_row
        ):
            with self.assertRaises(TypeError):
                self.t_env.from_elements(elements, schema)


class VectorUDT(UserDefinedType):

    @classmethod
    def sql_type(cls):
        return DataTypes.ROW(
            [
                DataTypes.FIELD("type", DataTypes.TINYINT()),
                DataTypes.FIELD("size", DataTypes.INT()),
                DataTypes.FIELD("indices", DataTypes.ARRAY(DataTypes.INT())),
                DataTypes.FIELD("values", DataTypes.ARRAY(DataTypes.DOUBLE())),
            ]
        )

    @classmethod
    def module(cls):
        return "pyflink.ml.core.linalg"

    def serialize(self, obj):
        if isinstance(obj, DenseVector):
            values = [float(v) for v in obj._values]
            return 1, None, None, values
        else:
            raise TypeError("Cannot serialize {!r} of type {!r}".format(obj, type(obj)))

    def deserialize(self, datum):
        pass


class DenseVector(object):
    __UDT__ = VectorUDT()

    def __init__(self, values):
        self._values = values

    def size(self) -> int:
        return len(self._values)

    def get(self, i: int):
        return self._values[i]

    def to_array(self):
        return self._values

    @property
    def values(self):
        return self._values

    def __str__(self):
        return "[" + ",".join([str(v) for v in self._values]) + "]"

    def __repr__(self):
        return "DenseVector([%s])" % (", ".join(str(i) for i in self._values))


class MyTimestampAssigner(TimestampAssigner):

    def extract_timestamp(self, value, record_timestamp) -> int:
        return int(value[0])


class MyTumblingEventTimeWindow(MergingWindowAssigner[tuple, TimeWindow]):

    def merge_windows(self,
                      windows,
                      callback: 'MergingWindowAssigner.MergeCallback[TimeWindow]') -> None:
        window_list = [w for w in windows]
        window_list.sort()
        for i in range(1, len(window_list)):
            if window_list[i - 1].end > window_list[i].start:
                callback.merge([window_list[i - 1], window_list[i]],
                               TimeWindow(window_list[i - 1].start, window_list[i].end))

    def assign_windows(self,
                       element: tuple,
                       timestamp: int,
                       context):
        return [TimeWindow(timestamp, timestamp + 5)]

    def get_default_trigger(self, env) -> Trigger[tuple, TimeWindow]:
        return SimpleTimeWindowTrigger()

    def get_window_serializer(self) -> TypeSerializer[TimeWindow]:
        return TimeWindowSerializer()

    def is_event_time(self) -> bool:
        return True


class SimpleTimeWindowTrigger(Trigger[tuple, TimeWindow]):

    def on_element(self,
                   element: tuple,
                   timestamp: int,
                   window: TimeWindow,
                   ctx: 'Trigger.TriggerContext') -> TriggerResult:
        return TriggerResult.CONTINUE

    def on_processing_time(self,
                           time: int,
                           window: TimeWindow,
                           ctx: 'Trigger.TriggerContext') -> TriggerResult:
        return TriggerResult.CONTINUE

    def on_event_time(self,
                      time: int,
                      window: TimeWindow,
                      ctx: 'Trigger.TriggerContext') -> TriggerResult:
        if time >= window.max_timestamp():
            return TriggerResult.FIRE_AND_PURGE
        else:
            return TriggerResult.CONTINUE

    def on_merge(self,
                 window: TimeWindow,
                 ctx: 'Trigger.OnMergeContext') -> None:
        pass

    def clear(self,
              window: TimeWindow,
              ctx: 'Trigger.TriggerContext') -> None:
        pass


class SumWindowFunction(WindowFunction[tuple, tuple, str, TimeWindow]):

    def apply(self, key: str, window: TimeWindow, inputs: Iterable[tuple]):
        result = 0
        for i in inputs:
            result += i[1]
        return [(key, result)]



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from typing import Dict, Set, Union

from py4j.java_gateway import JavaObject

from pyflink.common import DeserializationSchema
from pyflink.datastream.connectors import Source
from pyflink.datastream.connectors.kafka import KafkaOffsetsInitializer
from pyflink.java_gateway import get_gateway

__all__ = [
    'DynamicKafkaSource',
    'DynamicKafkaSourceBuilder',
    'KafkaMetadataService',
    'KafkaRecordDeserializationSchema',
    'KafkaStreamSetSubscriber',
    'KafkaStreamSubscriber',
    'StreamPatternSubscriber',
    'SingleClusterTopicMetadataService'
]


class KafkaMetadataService(object):
    """
    Base class for Kafka metadata service wrappers.
    """

    def __init__(self, j_metadata_service: JavaObject):
        self._j_metadata_service = j_metadata_service


class SingleClusterTopicMetadataService(KafkaMetadataService):
    """
    A KafkaMetadataService backed by a single Kafka cluster where stream ids map to topics.
    """

    def __init__(self, kafka_cluster_id: str, properties: Dict[str, str]):
        gateway = get_gateway()
        j_properties = gateway.jvm.java.util.Properties()
        for key, value in properties.items():
            j_properties.setProperty(key, value)
        j_service = gateway.jvm.org.apache.flink.connector.kafka.dynamic.metadata \
            .SingleClusterTopicMetadataService(kafka_cluster_id, j_properties)
        super().__init__(j_service)


class KafkaStreamSubscriber(object):
    """
    Wrapper for Java KafkaStreamSubscriber implementations.
    """

    def __init__(self, j_kafka_stream_subscriber: JavaObject):
        self._j_kafka_stream_subscriber = j_kafka_stream_subscriber


class KafkaStreamSetSubscriber(KafkaStreamSubscriber):
    """
    Subscriber that consumes from a fixed set of stream ids.
    """

    def __init__(self, stream_ids: Set[str]):
        gateway = get_gateway()
        j_stream_ids = gateway.jvm.java.util.HashSet()
        for stream_id in stream_ids:
            j_stream_ids.add(stream_id)
        j_subscriber = gateway.jvm.org.apache.flink.connector.kafka.dynamic.source.enumerator \
            .subscriber.KafkaStreamSetSubscriber(j_stream_ids)
        super().__init__(j_subscriber)


class StreamPatternSubscriber(KafkaStreamSubscriber):
    """
    Subscriber that consumes from stream ids matching a regex pattern.
    """

    def __init__(self, stream_pattern: str):
        gateway = get_gateway()
        j_pattern = gateway.jvm.java.util.regex.Pattern.compile(stream_pattern)
        j_subscriber = gateway.jvm.org.apache.flink.connector.kafka.dynamic.source.enumerator \
            .subscriber.StreamPatternSubscriber(j_pattern)
        super().__init__(j_subscriber)


class KafkaRecordDeserializationSchema(DeserializationSchema):
    """
    Wrapper for KafkaRecordDeserializationSchema.
    """

    def __init__(self, j_deserialization_schema: JavaObject):
        super().__init__(j_deserialization_schema)

    @staticmethod
    def value_only(deserialization_schema: DeserializationSchema) -> \
            'KafkaRecordDeserializationSchema':
        jvm = get_gateway().jvm
        j_deserializer = jvm.org.apache.flink.connector.kafka.source.reader.deserializer \
            .KafkaRecordDeserializationSchema.valueOnly(
                deserialization_schema._j_deserialization_schema)
        return KafkaRecordDeserializationSchema(j_deserializer)


class DynamicKafkaSource(Source):
    """
    Source implementation for dynamic Kafka streams.

    Example:
    ::

        >>> metadata_service = SingleClusterTopicMetadataService(
        ...     'cluster-a', {'bootstrap.servers': 'localhost:9092'})
        >>> source = DynamicKafkaSource.builder() \\
        ...     .set_stream_ids({'stream-a'}) \\
        ...     .set_kafka_metadata_service(metadata_service) \\
        ...     .set_value_only_deserializer(SimpleStringSchema()) \\
        ...     .build()
    """

    def __init__(self, j_dynamic_kafka_source: JavaObject):
        super().__init__(j_dynamic_kafka_source)

    @staticmethod
    def builder() -> 'DynamicKafkaSourceBuilder':
        return DynamicKafkaSourceBuilder()


class DynamicKafkaSourceBuilder(object):
    """
    Builder for DynamicKafkaSource.
    """

    def __init__(self):
        self._j_builder = get_gateway().jvm.org.apache.flink.connector.kafka.dynamic.source \
            .DynamicKafkaSource.builder()

    def build(self) -> DynamicKafkaSource:
        return DynamicKafkaSource(self._j_builder.build())

    def set_stream_ids(self, stream_ids: Set[str]) -> 'DynamicKafkaSourceBuilder':
        """
        Set the stream ids to consume.
        """
        j_set = get_gateway().jvm.java.util.HashSet()
        for stream_id in stream_ids:
            j_set.add(stream_id)
        self._j_builder.setStreamIds(j_set)
        return self

    def set_stream_pattern(self, stream_pattern: str) -> 'DynamicKafkaSourceBuilder':
        """
        Set a regex pattern to match stream ids.
        """
        self._j_builder.setStreamPattern(get_gateway().jvm.java.util.regex
                                         .Pattern.compile(stream_pattern))
        return self

    def set_kafka_stream_subscriber(
            self,
            kafka_stream_subscriber: Union[KafkaStreamSubscriber, JavaObject]) \
            -> 'DynamicKafkaSourceBuilder':
        """
        Set a custom KafkaStreamSubscriber.
        """
        if isinstance(kafka_stream_subscriber, KafkaStreamSubscriber):
            j_subscriber = kafka_stream_subscriber._j_kafka_stream_subscriber
        else:
            j_subscriber = kafka_stream_subscriber
        self._j_builder.setKafkaStreamSubscriber(j_subscriber)
        return self

    def set_kafka_metadata_service(
            self,
            kafka_metadata_service: Union[KafkaMetadataService, JavaObject]) \
            -> 'DynamicKafkaSourceBuilder':
        """
        Set the KafkaMetadataService.
        """
        if isinstance(kafka_metadata_service, KafkaMetadataService):
            j_metadata_service = kafka_metadata_service._j_metadata_service
        else:
            j_metadata_service = kafka_metadata_service
        self._j_builder.setKafkaMetadataService(j_metadata_service)
        return self

    def set_deserializer(self,
                         deserializer: Union[KafkaRecordDeserializationSchema, JavaObject]) \
            -> 'DynamicKafkaSourceBuilder':
        """
        Set the KafkaRecordDeserializationSchema.
        """
        if isinstance(deserializer, KafkaRecordDeserializationSchema):
            j_deserializer = deserializer._j_deserialization_schema
        else:
            j_deserializer = deserializer
        self._j_builder.setDeserializer(j_deserializer)
        return self

    def set_value_only_deserializer(self, deserialization_schema: DeserializationSchema) \
            -> 'DynamicKafkaSourceBuilder':
        """
        Set a value-only DeserializationSchema.
        """
        return self.set_deserializer(
            KafkaRecordDeserializationSchema.value_only(deserialization_schema))

    def set_starting_offsets(self, starting_offsets_initializer: KafkaOffsetsInitializer) \
            -> 'DynamicKafkaSourceBuilder':
        """
        Set the starting offsets for all streams.
        """
        self._j_builder.setStartingOffsets(starting_offsets_initializer._j_initializer)
        return self

    def set_bounded(self, stopping_offsets_initializer: KafkaOffsetsInitializer) \
            -> 'DynamicKafkaSourceBuilder':
        """
        Set the source to bounded mode with stopping offsets.
        """
        self._j_builder.setBounded(stopping_offsets_initializer._j_initializer)
        return self

    def set_properties(self, props: Dict[str, str]) -> 'DynamicKafkaSourceBuilder':
        """
        Set consumer properties for all clusters.
        """
        gateway = get_gateway()
        j_properties = gateway.jvm.java.util.Properties()
        for key, value in props.items():
            j_properties.setProperty(key, value)
        self._j_builder.setProperties(j_properties)
        return self

    def set_property(self, key: str, value: str) -> 'DynamicKafkaSourceBuilder':
        """
        Set a consumer property for all clusters.
        """
        self._j_builder.setProperty(key, value)
        return self

    def set_group_id(self, group_id: str) -> 'DynamicKafkaSourceBuilder':
        """
        Set the consumer group id for all clusters.
        """
        self._j_builder.setGroupId(group_id)
        return self

    def set_client_id_prefix(self, prefix: str) -> 'DynamicKafkaSourceBuilder':
        """
        Set the client id prefix for all clusters.
        """
        self._j_builder.setClientIdPrefix(prefix)
        return self
